/*
 * Copyright 1999-2004 The Apache Software Foundation.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/* $Id$ */

package org.apache.fop.render.afp;

import java.io.IOException;
import java.io.OutputStream;
import java.io.UnsupportedEncodingException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.fop.apps.FOPException;
import org.apache.fop.fo.Property;
import org.apache.fop.image.FopImage;
import org.apache.fop.layout.Area;
import org.apache.fop.layout.BlockArea;
import org.apache.fop.layout.BorderAndPadding;
import org.apache.fop.layout.FontInfo;
import org.apache.fop.layout.FontState;
import org.apache.fop.layout.Page;
import org.apache.fop.layout.inline.ForeignObjectArea;
import org.apache.fop.layout.inline.WordArea;
import org.apache.fop.pdf.PDFColor;
import org.apache.fop.pdf.PDFPathPaint;
import org.apache.fop.render.PrintRenderer;
import org.apache.fop.render.afp.extensions.AFPElement;
import org.apache.fop.render.afp.fonts.FontCollection;
import org.apache.fop.render.afp.fonts.FontFactory;
import org.apache.fop.render.afp.fonts.FontSetup;
import org.apache.fop.render.afp.modca.AFPConstants;
import org.apache.fop.render.afp.modca.AFPDataStream;
import org.apache.fop.render.afp.modca.IndexingException;
import org.apache.fop.svg.SVGArea;

/**
 * This is an implementaion of an FOP Renderer that renders areas to AFP.
 * <p>
 * A renderer is primarily designed to convert a given area tree into the output
 * document format. It should be able to produce pages and fill the pages with 
 * the text and graphical content. Usually the output is sent to an output 
 * stream. Some output formats may support extra information that is not 
 * available from the area tree or depends on the destination of the document.
 * Each renderer is given an area tree to render to its output format. The area 
 * tree is simply a representation of the pages and the placement of text and 
 * graphical objects on those pages.
 * </p><p>
 * The renderer will be given each page as it is ready and an output stream 
 * to write the data out. All pages are supplied in the order they appear in 
 * the document. In order to save memory it is possble to render the pages 
 * out of order. Any page that is not ready to be rendered is setup by the 
 * renderer first so that it can reserve a space or reference for when the 
 * page is ready to be rendered.The renderer is responsible for managing the 
 * output format and associated data and flow.
 * </p><p>
 * Each renderer is totally responsible for its output format. Because font 
 * metrics (and therefore layout) are obtained in two different ways depending
 * on the renderer, the renderer actually sets up the fonts being used. The 
 * font metrics are used during the layout process to determine the size of 
 * characters. 
 * </p><p>
 * The render context is used by handlers. It contains information about the 
 * current state of the renderer, such as the page, the position, and any other
 * miscellanous objects that are required to draw into the page.
 * </p><p>
 * A renderer is created by implementing the Renderer interface. However, the 
 * AbstractRenderer does most of what is needed, including iterating through 
 * the tree parts, so it is this that is extended. This means that this object
 * only need to implement the basic functionality such as text, images, and 
 * lines. AbstractRenderer's methods can easily be overridden to handle things
 * in a different way or do some extra processing.
 * </p><p>
 * The relevent AreaTree structures that will need to be rendered are Page, 
 * Viewport, Region, Span, Block, Line, Inline. A renderer implementation 
 * renders each individual page, clips and aligns child areas to a viewport,
 * handle all types of inline area, text, image etc and draws various lines 
 * and rectangles.
 * </p>
 * @author Pete Townsend <a href="mailto:pete@townsend.uk.com"/>
 * 
 * Note: There are specific extensions that have been added to the
 * FO. They are specific to there location within the FO and have to be 
 * processed accordingly (ie. at the start or end of the page).
 */
public class AFPRenderer extends PrintRenderer {

	/**
	 * Static logging instance
	 */
	protected static final Log LOGGER =
		LogFactory.getLog("org.apache.fop.render.afp");

	/**
	 * AFP factor for a 240 resolution = 72000/240 = 300
	 */
	private static final int DPI_CONVERSION_FACTOR_240 = 300;

	/**
	 * AFP factor for a 300 resolution = 72000/300 = 240
	 */
	private static final int DPI_CONVERSION_FACTOR_300 = 240;

	/**
	 * The AFP conversion factor to convert millipoints to AFP pels
	 */
	private int _dpiConversionFactor = DPI_CONVERSION_FACTOR_240;

	/**
	 * The afp data stream object responsible for generating afp data
	 */
	private AFPDataStream _afpDataStream = null;

	/**
	 * The fonts on the current page
	 */
	private HashMap _currentPageFonts = null;

	/**
	 * The current font color object
	 */
	private AFPFontColor _currentFontColor = null;

	/**
	 * The page font number counter, used to determine the next font reference
	 */
	private int _pageFontCounter = 0;
	
	/**
	 * The current font family
	 */
	private String _currentFontFamily = "";

	/**
	 * The Options to be set on the AFPRenderer
	 */
	private Map _afpOptions = null;

	/**
	 * The current page number
	 */
	private int _currentPageNumber = 0;

	/**
	 * The page width
	 */
	private int _pageWidth = 0;

	/**
	 * The page height
	 */
	private int _pageHeight = 0;

	/**
	 * The dpi resolution 
	 */
	private int _dpi = 0;

	/**
	 * The page extensions
	 */
	private List _extensions = null;

	/**
	 * Constructor for AFPRenderer, uses the default dpi resolution.
	 */
	public AFPRenderer() {
		this(240);
	}

	/**
	 * Constructor for AFPRenderer.
	 */
	public AFPRenderer(int dpi) {
		super();

		_dpi = dpi;

		if (dpi == 240) {
			_dpiConversionFactor = DPI_CONVERSION_FACTOR_240;
		} else if (dpi == 300) {
			_dpiConversionFactor = DPI_CONVERSION_FACTOR_300;
		} else {
			String msg = "DPI resolution " + dpi + " not supported";
			LOGGER.error(msg);
			throw new IllegalArgumentException(msg);
		}
	}

	/**
	  * Add a line to the current stream.
	  * @param x1 the start x location in millipoints
	  * @param y1 the start y location in millipoints
	  * @param x2 the end x location in millipoints
	  * @param y2 the end y location in millipoints
	  * @param th the thickness in millipoints
	  * @param rs the rule style
	  * @param stroke the stroke color/gradient
	  */
	protected void addLine(
		int x1,
		int y1,
		int x2,
		int y2,
		int th,
		int rs,
		PDFPathPaint stroke) {

		// Ignore the rule style as AFP only supprts normal lines
		addLine(x1, y1, x2, y2, th, stroke);

	}

	/**
	  * Add a line to the current stream.
	  * @param x1 the start x location in millipoints
	  * @param y1 the start y location in millipoints
	  * @param x2 the end x location in millipoints
	  * @param y2 the end y location in millipoints
	  * @param th the thickness in millipoints
	  * @param stroke the stroke color/gradient
	  */
	protected void addLine(
		int x1,
		int y1,
		int x2,
		int y2,
		int th,
		PDFPathPaint stroke) {

		_afpDataStream.createLine(
			(x1 / _dpiConversionFactor),
			(_pageHeight - (y1 / _dpiConversionFactor)),
			(x2 / _dpiConversionFactor),
			(_pageHeight - (y2 / _dpiConversionFactor)),
			(Math.abs(th) / _dpiConversionFactor));

	}

	/**
	 * @see org.apache.fop.render.PrintRenderer#addRect(int, int, int, int, PDFPathPaint, PDFPathPaint)
	 */
	protected void addRect(
		int x,
		int y,
		int w,
		int h,
		PDFPathPaint stroke,
		PDFPathPaint fill) {

		// Assume fill is grey by default
		int red = 105;
		int green = 105;
		int blue = 105;

		// Fill should be an instance of PDFColor, so obtain the correct RGB
		if (fill instanceof PDFColor) {
			PDFColor color = (PDFColor) fill;
			red = color.red255();
			green = color.green255();
			blue = color.blue255();
		}

		// Create the raster data for the color shade requested.
		_afpDataStream.createShading(
			x / _dpiConversionFactor,
			_pageHeight - (y / _dpiConversionFactor),
			w / _dpiConversionFactor,
			(-h / _dpiConversionFactor),
			red,
			green,
			blue);

		// Create the rectangle boundaries
		addRect(x, y, w, h, stroke);

	}

	/**
	 * @see org.apache.fop.render.PrintRenderer#addRect(int, int, int, int, PDFPathPaint)
	 */
	protected void addRect(int x, int y, int w, int h, PDFPathPaint stroke) {

		int thickness = 1000;
		//Draw the top line
		addLine(x, y, x + w, y, thickness, stroke);

		//Draw the left line down
		addLine(x, y, x, y + h, thickness, stroke);

		//Draw the bottom line
		addLine(x + w, y + h, x, y + h, thickness, stroke);

		//Draw the right line up
		addLine(x + w, y + h, x + w, y, thickness, stroke);

	}

	/**
	 * @see org.apache.fop.render.Renderer#renderForeignObjectArea(ForeignObjectArea)
	 */
	public void renderForeignObjectArea(ForeignObjectArea area) {

		//TODO

	}

	/**
	 * @see org.apache.fop.render.Renderer#renderPage(Page)
	 */
	public void renderPage(Page page) {

		processStartDocExtensions(page);

		_currentFontFamily = "";
		super.currentFontSize = 0;
		_pageFontCounter = 0;
		_currentPageFonts.clear();

		_pageWidth =
			(int) ((float) page.getWidth() / _dpiConversionFactor + .5);
		_pageHeight =
			(int) ((float) page.getHeight() / _dpiConversionFactor + .5);

		_afpDataStream.startPage(_pageWidth, _pageHeight);

		//now do the start of page extensions
		processStartPageExtensions(page);

		renderRegions(page);

		Iterator i = _currentPageFonts.values().iterator();
		while (i.hasNext()) {
			AFPFontAttributes afpFontAttributes = (AFPFontAttributes) i.next();

			_afpDataStream.createFont(
				afpFontAttributes.getFontReference(),
				afpFontAttributes.getFontFamily(),
				afpFontAttributes.getFontWeight(),
				afpFontAttributes.getFontStyle(),
				afpFontAttributes.getPointSize());

		}

		_afpDataStream.endPage();

		/**
		 * After the current page is ended, get the page extenstions
		 * and process any extension tags that may need to be added to
		 * either the document or the current page.
		 */
		processPageEndExtensions(page);

		// ensures that color is properly reset
		super.currentFill = null;

	}

	/**
	 * @see org.apache.fop.render.Renderer#renderSVGArea(SVGArea)
	 */
	public void renderSVGArea(SVGArea area) {

		//TODO

	}

	/**
	 * @see org.apache.fop.render.Renderer#renderWordArea(WordArea)
	 */
	public void renderWordArea(WordArea area) {

		FontState fontState = area.getFontState();
		// Get font details from the WordArea font state object
		_currentFontFamily = fontState.getFontFamily();		
		
		super.currentFontSize = fontState.getFontSize();
		String fontWeight = fontState.getFontWeight();
		String fontStyle = fontState.getFontStyle();
		
		int pointSize = super.currentFontSize / 1000;
		boolean isUnderlined = area.getUnderlined();

		_currentFontColor.setRed(area.getRed());
		_currentFontColor.setGeen(area.getGreen());
		_currentFontColor.setBlue(area.getBlue());

		PDFColor areaColor = null;
		if (this.currentFill instanceof PDFColor) {
			areaColor = (PDFColor) this.currentFill;
		}

		if (areaColor == null
			|| areaColor.red() != (double) _currentFontColor.getRed()
			|| areaColor.green() != (double) _currentFontColor.getGreen()
			|| areaColor.blue() != (double) _currentFontColor.getBlue()) {

			areaColor =
				new PDFColor(
					(double) _currentFontColor.getRed(),
					(double) _currentFontColor.getGreen(),
					(double) _currentFontColor.getBlue());

			this.currentFill = areaColor;

		}

		int xpos = super.currentXPosition;
		int ypos = super.currentYPosition;

		super.addWordLines(area, xpos, ypos, super.currentFontSize, areaColor);

		String worddata = area.getText();
		
		FontInfo fontInfo = fontState.getFontInfo();
		if (!fontInfo.hasFont(_currentFontFamily, fontStyle, fontWeight)) {
			LOGGER.error("unknown font " + _currentFontFamily
											   + " so defaulted font to any");
			_currentFontFamily = "any";
		}
		
		// Create an AFPFontAttributes object from the current font details
		AFPFontAttributes afpFontAttributes =
			new AFPFontAttributes(
				_currentFontFamily,
				fontWeight,
				fontStyle,
				pointSize);

		if (!_currentPageFonts.containsKey(afpFontAttributes.getFontKey())) {

			// Font not found on current page, so add the new one
			_pageFontCounter++;
			afpFontAttributes.setFontReference(_pageFontCounter);
			_currentPageFonts.put(
				afpFontAttributes.getFontKey(),
				afpFontAttributes);

		} else {

			// Use the previously stored font attributes
			afpFontAttributes =
				(AFPFontAttributes) _currentPageFonts.get(
					afpFontAttributes.getFontKey());

		}

		// Try and get the encoding to use for the font
		String encoding = null;

		FontFactory fontFactory = FontFactory.getInstance();
		FontCollection fontCollection = fontFactory.getFontCollection(_dpi);

		try {
			encoding =
				fontCollection
					.getFont(afpFontAttributes.getFontFamily())
					.getEncoding();
		} catch (Throwable ex) {
			encoding = AFPConstants.EBCIDIC_ENCODING;
			LOGGER.warn(
				"renderWordArea():: Error getting encoding for font family "
					+ afpFontAttributes.getFontFamily()
					+ " - using default encoding "
					+ encoding);
		}

		try {

			_afpDataStream.createText(
				afpFontAttributes.getFontReference(),
				(int) ((xpos / _dpiConversionFactor)),
				(int) (_pageHeight - ypos / _dpiConversionFactor),
				worddata.getBytes(encoding));

		} catch (UnsupportedEncodingException usee) {

			LOGGER.error(
				"renderWordArea:: Font family "
					+ afpFontAttributes.getFontFamily()
					+ " caused UnsupportedEncodingException");

		}

		super.currentXPosition += (int) area.getContentWidth();

	}

	/**
	 * @see org.apache.fop.render.Renderer#setProducer(String)
	 */
	public void setProducer(String producer) {

		_afpDataStream.setProducer(producer);

	}

	/**
	 * @see org.apache.fop.render.AbstractRenderer#drawImageClipped(int, int, int, int, int, int, FopImage, FontState)
	 */
	protected void drawImageClipped(
		int x,
		int y,
		int clipX,
		int clipY,
		int clipW,
		int clipH,
		FopImage image,
		FontState fs) {

		//TODO

	}
	/**
	 * Override the doFrame(Area area) method so that we draw
	 * lines rather than rectangles when processing an area.
	 */
	protected void doFrame(Area area) {
		int w, h;
		int rx = this.currentAreaContainerXPosition;
		w = area.getContentWidth();
		if (area instanceof BlockArea)
			rx += ((BlockArea) area).getStartIndent();
		h = area.getContentHeight();
		int ry = this.currentYPosition;

		rx = rx - area.getPaddingLeft();
		ry = ry + area.getPaddingTop();
		w = w + area.getPaddingLeft() + area.getPaddingRight();
		h = h + area.getPaddingTop() + area.getPaddingBottom();

		doBackground(area, rx, ry, w, h);

		rx = rx - area.getBorderLeftWidth();
		ry = ry + area.getBorderTopWidth();
		w = w + area.getBorderLeftWidth() + area.getBorderRightWidth();
		h = h + area.getBorderTopWidth() + area.getBorderBottomWidth();

		// Handle line style
		// Offset for haft the line width!
		BorderAndPadding bp = area.getBorderAndPadding();
		int left = rx - area.getBorderLeftWidth() / 2;
		int right = rx + w + area.getBorderRightWidth() / 2;
		int top = ry + area.getBorderTopWidth() / 2;
		int bottom = ry - h - area.getBorderBottomWidth() / 2;
		if (area.getBorderTopWidth() != 0)
			addLine(
				left,
				top,
				right,
				top,
				area.getBorderTopWidth(),
				new PDFColor(bp.getBorderColor(BorderAndPadding.TOP)));
		if (area.getBorderLeftWidth() != 0)
			addLine(
				left,
				ry + area.getBorderTopWidth(),
				left,
				bottom,
				area.getBorderLeftWidth(),
				new PDFColor(bp.getBorderColor(BorderAndPadding.LEFT)));
		if (area.getBorderRightWidth() != 0)
			addLine(
				right,
				ry + area.getBorderTopWidth(),
				right,
				bottom,
				area.getBorderRightWidth(),
				new PDFColor(bp.getBorderColor(BorderAndPadding.RIGHT)));
		if (area.getBorderBottomWidth() != 0)
			addLine(
				rx - area.getBorderLeftWidth(),
				bottom,
				rx + w + area.getBorderRightWidth(),
				bottom,
				area.getBorderBottomWidth(),
				new PDFColor(bp.getBorderColor(BorderAndPadding.BOTTOM)));

	}

	/**
	 * @see org.apache.fop.render.AbstractRenderer#drawImageScaled(int, int, int, int, FopImage, FontState)
	 */
	protected void drawImageScaled(
		int x,
		int y,
		int w,
		int h,
		FopImage image,
		FontState fs) {

		//TODO

	}

	/**
	 * @see org.apache.fop.render.Renderer#render(Page, OutputStream)
	 */
	public void render(Page page, OutputStream stream) throws IOException {

		_currentPageNumber++;

		LOGGER.debug(
			"staring rendering of page number "
				+ _currentPageNumber
				+ " to AFP");

		super.idReferences = page.getIDReferences();

		renderPage(page);

		LOGGER.debug(
			"finished rendering page number " + _currentPageNumber + " to AFP");

	}

	/**
	 * @see org.apache.fop.render.Renderer#startRenderer(java.io.OutputStream)
	 */
	public void startRenderer(OutputStream outputStream) throws IOException {

		_currentPageFonts = new HashMap();
		_currentFontColor = new AFPFontColor(0, 0, 0);
		_currentPageNumber = 0;
		_afpDataStream = new AFPDataStream(_dpi);
		_afpDataStream.startDocument();
		super.startRenderer(outputStream);

	}

	/**
	 * @see org.apache.fop.render.Renderer#stopRenderer(java.io.OutputStream)
	 */
	public void stopRenderer(OutputStream outputStream) throws IOException {

		_afpDataStream.endDocument();
		super.stopRenderer(outputStream);
		_afpDataStream.write(outputStream);
		outputStream.flush();

	}

	/**
	 * @see org.apache.fop.render.Renderer#setOptions(Map)
	 */
	public void setOptions(Map options) {

		_afpOptions = options;

	}

	/**
	 * set up the font info
	 * @param fontInfo font info to set up
	 */
	public void setupFontInfo(FontInfo fontInfo) throws FOPException {
		this.fontInfo = fontInfo;
		FontSetup.setup(_dpi, fontInfo);
	}

	/**
	 * Method to handle the start of page extensions. These extension provide
	 * for the inclusion of non-standard components in the rendering of the 
	 * formatted object into an AFP data stream. 
	 * @param page the current page object.
	 */
	protected void processStartPageExtensions(Page page) {

		_extensions = page.getExtensions();

		if (_extensions == null) {
			// No extensions to process
			return;
		}

		Iterator iterator = _extensions.iterator();

		while (iterator.hasNext()) {

			Object obj = iterator.next();
			if (obj instanceof AFPElement) {
				AFPElement afpElement = (AFPElement) obj;
				String nameOfElement = afpElement.getName();
				LOGGER.debug(
					"Processing AFP extension element: " + nameOfElement);

				if (nameOfElement.equals("afp:overlay")) {

					Property name = afpElement.getProperty("name");
					Property or = afpElement.getProperty("orientation");
					renderStaticOverlay(name, or);

				}

				if (nameOfElement.equals("afp:dynamic-overlay")) {
					Property name = afpElement.getProperty("name");
					Property xCoor = afpElement.getProperty("x-axis");
					Property yCoor = afpElement.getProperty("y-axis");
					Property or = afpElement.getProperty("orientation");
					renderDynamicOverlay(name, xCoor, yCoor, or);
				}

			}
		}
	}

	/**
	 * Method to handle the start of document extensions. These extension 
	 * provide for the inclusion of non-standard components in the rendering 
	 * of the formatted object into an AFP data stream. 
	 * @param page the current page object.
	 */
	protected void processStartDocExtensions(Page page) {

		_extensions = page.getExtensions();

		if (_extensions == null) {
			// No extensions to process
			return;
		}

		Iterator iterator = _extensions.iterator();

		while (iterator.hasNext()) {

			Object obj = iterator.next();
			if (obj instanceof AFPElement) {
				AFPElement afpElement = (AFPElement) obj;
				String nameOfElement = afpElement.getName();
				LOGGER.debug(
					"Processing AFP extension element: " + nameOfElement);

				if (nameOfElement.equals("afp:page-group")) {
					String pageGroupName =
						afpElement.getProperty("name").getString();
					_afpDataStream.createPageGroup(pageGroupName);

					Collection kids = areChildren(afpElement);

					if (kids.isEmpty()) {
						return;
					} else {
						processIndexingTags(kids);
					}
				}
			}
		}
	}

	/**
	 * Method to handle the end of page extensions. These extension 
	 * provide for the inclusion of non-standard components in the rendering 
	 * of the formatted object into an AFP data stream. 
	 * @param page the current page object.
	 */
	protected void processPageEndExtensions(Page page) {

		_extensions = page.getExtensions();

		if (_extensions == null) {
			// No extensions to process
			return;
		}

		Iterator iterator = _extensions.iterator();

		while (iterator.hasNext()) {

			Object obj = iterator.next();
			if (obj instanceof AFPElement) {
				AFPElement afpElement = (AFPElement) obj;
				String nameOfElement = afpElement.getName();
				LOGGER.debug(
					"Processing AFP extension element: " + nameOfElement);
				//now do something with the actual extensions

				if (nameOfElement.equals("afp:end-page-group")) {
					LOGGER.debug(
						"Got a end-page-group extension, processing it");
					String endPageGroupName =
						afpElement.getProperty("name").getString();

					try {
						_afpDataStream.endPageGroup(endPageGroupName);
					} catch (IndexingException ie) {
						LOGGER.error("Error trying to index the doc");
						throw new IllegalStateException();
					}

				}

			}

		}
	}

	/**
	 * Deals with the specifics of rendering the custom afp:overlay
	 * tag. 
	 * Note: it creates two objects (behind the scene). This is an 
	 * IPO (IncludePageObject) and MPO (MapPageObject).
	 * @param name - name of the overlay
	 * @param orientation - orientation of the overlay
	 * 
	 */
	private void renderStaticOverlay(Property name, Property orientation) {

		//We are not going to know how many overlays are required in 
		//the page, as we are reading an element at a time. For this reason
		//we are going to have to add to the MPO as we go.

		if ((name == null) || (orientation == null)) {
			LOGGER.error(
				"Trying to render an overlay, one of the values is null ");
			throw new IllegalArgumentException();
		}
		String overLayName = name.getString();

		LOGGER.debug("Adding an overlay for: " + overLayName);
		int or = Integer.parseInt(orientation.getString());
		_afpDataStream.createStaticOverlay(overLayName, or);
	}
	/**
	 * Deals with the specifics of rendering the custom afp:dynamic-overlay
	 * tag.
	 * @param name - name of the overlay
	 * @param xCoor - the x coordinate to position the overlay
	 * @param yCoor - the y coordinate to position the overlay
	 * @param orientation - the orientation required for the overlay 
	 */
	private void renderDynamicOverlay(
		Property name,
		Property xCoord,
		Property yCoord,
		Property orientation) {

		if ((name == null)
			|| (xCoord == null)
			|| (yCoord == null)
			|| (orientation == null)) {
			LOGGER.error(
				"Trying to render an overlay, one of the values is null ");
			throw new IllegalArgumentException();
		}
		String overLayName = name.getString();

		double xVal = Double.parseDouble(xCoord.getString());
		double yVal = Double.parseDouble(yCoord.getString());
		double or = Double.parseDouble(orientation.getString());
		LOGGER.debug(
			"Adding a dynamic overlay for: "
				+ overLayName
				+ " Coor x: "
				+ xVal
				+ " Coor y: "
				+ yVal
				+ " orientation: "
				+ or);

		//now convert the coordinate
		int newXVal = convertCoordinate(xVal);
		int newYVal = convertCoordinate(yVal);

		_afpDataStream.createDynamicOverlay(overLayName, newXVal, newYVal);
	}

	/**
	 * This method converts the parameter value (cms) to the dpi values.
	 * @param coor the coordinate to be converted from FOP coordinates to AFP.
	 */
	private int convertCoordinate(double coor) {

		double convertVal = coor * 0.3937;
		return (int) ((convertVal * 72000) / _dpiConversionFactor + .5);

	}

	/**
	 * Helper create the TLE indexing details (see: TagLogicalElement)
	 * @param element - an AFPElement (used to wrap the afp extensions)
	 * @return String[] of indexing name/value combinations
	 * 
	 */
	private ArrayList iterateOnExtensions(AFPElement element)
		throws IndexingException {

		//the start tag is :<afp:page-group>
		//and there may be children below it (but there doesn't have to be):
		//The child of this is :<afp:tag-logical-element>
		//the actual element used for indexing, is the child of the above,
		//which is :<afp:attribute> (which you must have if you have a 
		//<afp:tag-logical-element>), and there could be a number (max??)
		//of attributes for indexing

		ArrayList listOfAttributes = new ArrayList();
		int cnt = 0;

		Collection children = element.getChildren();
		Iterator i = children.iterator();

		while (i.hasNext()) {

			Object obj = i.next();
			if (obj instanceof AFPElement) {
				//the child we expect to be:afp:tag-logical-element
				AFPElement afpElement = (AFPElement) obj;
				String nameOfElement = afpElement.getName();
				if (nameOfElement.equals("afp:tag-logical-element")) {
					listOfAttributes = iterateOnExtensions(afpElement);
				}
				if (nameOfElement.equals("afp:attribute")) {

					//change the value in the nameOfElement so the when
					//the method returns above it will not hit this again
					nameOfElement = "afp:tag-logical-element";

					//now get the properties associatted with indexing
					listOfAttributes.add(
						afpElement.getProperty("name").getString());
					listOfAttributes.add(
						afpElement.getProperty("value").getString());
					LOGGER.debug(
						"we have some indexing attributes, they are name: "
							+ afpElement.getProperty("name").getString()
							+ " value: "
							+ afpElement.getProperty("value").getString());
				}
			}
		}
		//just before you send it, check that you have an even number, else its corrupt.
		if (!((listOfAttributes.size() % 2) == 0)) {
			throw new IndexingException("Issue in the FO, the indexing tag -afp:attribute, must have a name and value for each tag");
		}

		return listOfAttributes;

	}

	/**
	 * Helper to see if there are any children under the custom tags,
	 * if there are return the child.
	 * @param element - the AFPElement we are currently working with.
	 */
	private Collection areChildren(AFPElement element) {

		Collection children = element.getChildren();
		return children;

	}
	/**
	 * This method is responsible for iterating through the
	 * relevant tags to add the indexing tags to the
	 * data stream.
	 * @param elements - this is a collection of AFPElements
	 * under the parent.
	 * 
	 */
	private void processIndexingTags(Collection elements) {

		//Can only have one child under afp:page-group, this
		// is the afp:tag-logical-element tag
		if (!(elements.size() == 1)) {
			LOGGER.error("There is more than one tag under afp:page-group");
		}
		Iterator i = elements.iterator();
		AFPElement element = (AFPElement) i.next();

		String nameOfElement = element.getName();
		LOGGER.debug("processing indexing tags for element: " + nameOfElement);
		if (nameOfElement.equals("afp:tag-logical-element")) {
			//now get the children
			Collection child = element.getChildren();
			Iterator i2 = child.iterator();

			ArrayList listOfAttributes = new ArrayList();

			while (i2.hasNext()) {

				Object obj = i2.next();
				if (obj instanceof AFPElement) {
					AFPElement afpElement = (AFPElement) obj;

					if (afpElement.getName().equals("afp:attribute")) {
						//now get the properties associatted with indexing, do this for each afp:attribute tag. 
						listOfAttributes.add(
							afpElement.getProperty("name").getString());
						listOfAttributes.add(
							afpElement.getProperty("value").getString());
						LOGGER.debug(
							"we have some indexing attributes, they are name: "
								+ afpElement.getProperty("name").getString()
								+ " value: "
								+ afpElement.getProperty("value").getString());
					}

				}
			}
			_afpDataStream.createPageGroupTLE(listOfAttributes);
		}
	}

}
