/*
 * Copyright 1999-2004 The Apache Software Foundation.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/* $Id$ */

package org.apache.fop.render.afp.fonts;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.fop.render.afp.exceptions.FontRuntimeException;
import org.apache.fop.render.afp.modca.AFPConstants;
import org.apache.fop.render.afp.tools.StructuredFieldReader;

/**
 * The AFPFontReader is responsible for reading the font attributes from
 * binary code page files and the character set metric files. In IBM font 
 * structure, a code page maps each character of text to the characters in
 * a character set. Each character is translated into a code point. When the 
 * character is printed, each code point is matched to a character ID on the 
 * code page specified. The character ID is then matched to the image (raster 
 * pattern or outline pattern) of the character in the character set specified. 
 * The image in the character set is the image that is printed in the document. 
 * To be a valid code page for a particular character set, all character IDs 
 * in the code page must be included in that character set.
 * <p/>
 * This class will read the font information from the binary code page files 
 * and character set metric files in order to determine the correct metrics
 * to use when rendering the formatted object.
 * <p/>
 * @author Pete Townsend <a href="mailto:pete@townsend.uk.com"/>
 */
public final class AFPFontReader {

	/**
	 * Static logging instance
	 */
	protected static final Log LOGGER =
		LogFactory.getLog("org.apache.fop.render.afp.fonts");

	/**
	 * 
	 */
	private static int FOPF = 300000;

	/**
	 * The code page file extenstion
	 */
	private static final String CODEPAGE_FILE_EXTENSION = "CDP";

	/**
	 * The encoding to use to convert from EBCIDIC to ASCII
	 */
	private static final String ASCII_ENCODING = "UTF8";

	/**
	 * The collection of code pages
	 */
	private static HashMap _codePages = new HashMap();

	/**
	 * Load the font details and metrics into the CharacterSetMetric object, this
	 * will use the actual afp code page and character set files to load the object
	 * with the necessary metrics.	 
	 * @param characterSetMetric the CharacterSetMetric object to populate
	 */
	public static void loadCharacterSetMetric(CharacterSetMetric characterSetMetric) {

		InputStream inputStream = null;

		try {

			/**
			 * Get the code page which contains the character mapping 
			 * inormation to map the unicode character id to the 
			 * graphic chracter global identifier.
			 */
			String cp = new String(characterSetMetric.getCodePage());
			String path = characterSetMetric.getPath();

			HashMap codepage = (HashMap) _codePages.get(cp);

			if (codepage == null) {
				codepage =
					loadCodePage(cp, characterSetMetric.getEncoding(), path);
				_codePages.put(cp, codepage);
			}

			/**
			 * Load the character set metric information, no need to cache
			 * this information as it should be cached by the objects
			 * that wish to load character set metric information.
			 */
			StringBuffer buf = new StringBuffer();
			buf.append(characterSetMetric.getCharacterSet());
			buf.append(".");
			buf.append(characterSetMetric.getDPI());
			String fontResource = buf.toString();

			ClassLoader classLoader =
				Thread.currentThread().getContextClassLoader();
			if (classLoader == null) {
				classLoader = AFPFontReader.class.getClassLoader();
			}

			inputStream =
				classLoader.getResourceAsStream(
					path + File.separator + fontResource);
			if (inputStream == null) {
				throw new FontRuntimeException(
					"Failed to open resource " + fontResource);
			}

			StructuredFieldReader sfr = new StructuredFieldReader(inputStream);

			//process D3AE89 Font Orientation
			CharacterSetOrientation[] csoArray = processFontOrientation(sfr);

			//process D3AC89 Font Position
			processFontPosition(sfr, csoArray, characterSetMetric.getDPI());

			//process D38C89 Font Index (per orientation)
			for (int i = 0; i < csoArray.length; i++) {
				processFontIndex(
					sfr,
					csoArray[i],
					codepage,
					characterSetMetric.getDPI());
				characterSetMetric.addCharacterSetOrientation(csoArray[i]);
			}

		} catch (Exception ex) {
			throw new FontRuntimeException(
				"Failed to load the character set metrics for code page "
					+ characterSetMetric.getCodePage(),
				ex);
		} finally {
			try {
				inputStream.close();
			} catch (Exception ex) {
				// Ignore
			}
		}

	}

	/**
	 * Load the code page information from the appropriate file. The file name 
	 * to load is determined by the code page name and the file extension 'CDP'.
	 * @param codePage the code page identifier
	 * @param encoding the encoding to use for the character decoding
	 */
	private static HashMap loadCodePage(
		String codePage,
		String encoding,
		String path)
		throws IOException, FileNotFoundException {

		// Create the HashMap to store code page information
		HashMap codepages = new HashMap();

		// Create the name of the code page file
		StringBuffer buf = new StringBuffer();
		buf.append(codePage.trim());
		buf.append(".");
		buf.append(CODEPAGE_FILE_EXTENSION);
		String fontResource = buf.toString();
		
		ClassLoader classLoader = Thread.currentThread().getContextClassLoader();
		if (classLoader == null) {
			classLoader = AFPFontReader.class.getClassLoader();
		}

		InputStream is = classLoader.getResourceAsStream(
				path + File.separator + fontResource);
		if (is == null) {
			String msg =
				"AFPFontReader:: loadCodePage(String):: code page file not found for "
					+ codePage;
			LOGGER.error(msg);
			throw new FileNotFoundException(msg);
		}

		byte[] codepageSF = new byte[] {(byte) 0xD3, (byte) 0xA8, (byte) 0x87 };
		StructuredFieldReader sfr = new StructuredFieldReader(is);
		byte[] data = sfr.getNext(codepageSF);

		String stringData = new String(data, AFPConstants.EBCIDIC_ENCODING);
		String actualCodePage = stringData.substring(3, 11);

		if (actualCodePage == null || !actualCodePage.equals(codePage)) {
			String msg =
				"loadCodePage(String):: requested code page "
					+ codePage
					+ " does not match file code page "
					+ actualCodePage;
			LOGGER.error(msg);
			throw new FontRuntimeException(msg);
		}

		byte[] characterTableSF =
			new byte[] {(byte) 0xD3, (byte) 0x8C, (byte) 0x87 };
		data = sfr.getNext(characterTableSF);

		int position = 0;
		byte[] gcgiBytes = new byte[8];
		byte[] charBytes = new byte[1];

		for (int index = 0; index < data.length; index++) {

			// Ignore bytes 0 - 2
			if (index > 2) {
				if (position < 8) {
					// Build the graphic character global identifier key
					gcgiBytes[position] = data[index];
					position++;
				} else if (position == 9) {
					position = 0;
					// Set the character
					charBytes[0] = data[index];
					String gcgiString =
						new String(gcgiBytes, AFPConstants.EBCIDIC_ENCODING);
					String charString = new String(charBytes, encoding);
					int value = charString.charAt(0);
					codepages.put(gcgiString, charString);
				} else {
					position++;
				}
			}
		}

		try {
			is.close();
		} catch (Exception ex) {
			// Ignore
		}

		return codepages;

	}

	/**
	 * Process the font orientation details from using the structured field reader.
	 * @param sfr the structured field reader 
	 */
	private static CharacterSetOrientation[] processFontOrientation(StructuredFieldReader sfr)
		throws IOException {

		byte[] fontOrientationSF =
			new byte[] {(byte) 0xD3, (byte) 0xAE, (byte) 0x89 };
		byte[] data = sfr.getNext(fontOrientationSF);

		int position = 0;
		byte[] fnoData = new byte[26];

		ArrayList orientations = new ArrayList();

		for (int index = 0; index < data.length; index++) {

			// Ignore bytes 0 - 2
			if (index > 2) {

				// Build the font orientation record
				fnoData[position] = data[index];
				position++;

				if (position == 26) {

					position = 0;

					int orientation = 0;

					switch (fnoData[2]) {
						case 0x00 :
							orientation = 0;
							break;
						case 0x2D :
							orientation = 90;
							break;
						case 0x5A :
							orientation = 180;
							break;
						case (byte) 0x87 :
							orientation = 270;
							break;
						default :
							System.out.println("ERROR: Oriantation");
					}

					CharacterSetOrientation cso =
						new CharacterSetOrientation(orientation);
					orientations.add(cso);

				}

			}
		}

		CharacterSetOrientation[] csoArray =
			new CharacterSetOrientation[orientations.size()];
		Iterator i = orientations.iterator();
		int count = 0;
		while (i.hasNext()) {
			csoArray[count] = (CharacterSetOrientation) i.next();
			count++;
		}

		return csoArray;

	}

	/**
	 * Populate the CharacterSetOrientation object in the suplied array
	 * with the font position details using the supplied structured field reader.
	 * @param sfr the structured field reader
	 * @param csoArray the array of CharacterSetOrientation objects
	 */
	private static void processFontPosition(
		StructuredFieldReader sfr,
		CharacterSetOrientation[] csoArray,
		int dpi)
		throws IOException {

		byte[] fontPositionSF =
			new byte[] {(byte) 0xD3, (byte) 0xAC, (byte) 0x89 };
		byte[] data = sfr.getNext(fontPositionSF);

		int position = 0;
		byte[] fpData = new byte[26];

		int csoIndex = 0;

		int fopFactor = FOPF;

		for (int index = 0; index < data.length; index++) {

			// Ignore bytes 0 - 2
			if (index > 2) {

				if (position < 22) {
					// Build the font orientation record
					fpData[position] = data[index];
				} else if (position == 22) {

					position = 0;

					CharacterSetOrientation cso = csoArray[csoIndex];

					int xHeight =
						((fpData[2] & 0xFF) << 8) + (fpData[3] & 0xFF);
					int capHeight =
						((fpData[4] & 0xFF) << 8) + (fpData[5] & 0xFF);
					int ascHeight =
						((fpData[6] & 0xFF) << 8) + (fpData[7] & 0xFF);
					int dscHeight =
						((fpData[8] & 0xFF) << 8) + (fpData[9] & 0xFF);

					dscHeight = dscHeight * -1;

					cso.setXHeight(xHeight * fopFactor);
					cso.setCapHeight(capHeight * fopFactor);
					cso.setAscender(ascHeight * fopFactor);
					cso.setDescender(dscHeight * fopFactor);

					csoIndex++;

					fpData[position] = data[index];

				}

				position++;

			}
		}

	}

	/**
	 * Process the font index details for the character set orientation.
	 * @param sfr the structured field reader
	 * @param cso the CharacterSetOrientation object to populate
	 * @param codepage the map of code pages
	 */
	private static void processFontIndex(
		StructuredFieldReader sfr,
		CharacterSetOrientation cso,
		HashMap codepage,
		int dpi)
		throws IOException {

		byte[] fontIndexSF =
			new byte[] {(byte) 0xD3, (byte) 0x8C, (byte) 0x89 };
		byte[] data = sfr.getNext(fontIndexSF);

		int fopFactor = FOPF;
		int position = 0;

		byte[] gcgid = new byte[8];
		byte[] fiData = new byte[20];

		int lowest = 255;
		int highest = 0;

		for (int index = 0; index < data.length; index++) {

			// Ignore bytes 0 - 2
			if (index > 2) {

				if (position < 8) {
					gcgid[position] = (byte) data[index];
					position++;
				} else if (position < 27) {
					fiData[position - 8] = (byte) data[index];
					position++;
				} else if (position == 27) {

					fiData[position - 8] = (byte) data[index];

					position = 0;

					String gcgiString =
						new String(gcgid, AFPConstants.EBCIDIC_ENCODING);

					String idx = (String) codepage.get(gcgiString);

					if (idx != null) {

						int cidx = idx.charAt(0);
						int width =
							((fiData[0] & 0xFF) << 8) + (fiData[1] & 0xFF);

						if (cidx < lowest) {
							lowest = cidx;
						}

						if (cidx > highest) {
							highest = cidx;
						}

						int a = (width * fopFactor);

						cso.setWidth(cidx, a);

					}

				}

			}
		}

		cso.setFirstChar(lowest);
		cso.setLastChar(highest);

	}

}
