/*
 * Copyright 1999-2004 The Apache Software Foundation.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/* $Id$ */

package org.apache.fop.render.afp.fonts;

import java.util.Collection;
import java.util.HashMap;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.fop.layout.FontMetric;
import org.apache.fop.render.afp.exceptions.FontRuntimeException;

/**
 * A font is a set of graphic characters with similar design 
 * characteristics; that is, a font is a designer's concept of how a set of 
 * graphic characters should appear. Graphic characters (glyphs is the term
 * used by the ISO/IEC 9541 Font Information Interchange standard) are 
 * letters, numerals, punctuation marks, ideograms, or symbols that appear 
 * in text.
 * <p/>
 * Historically, a font was a collection of lead slugs containing the raised
 * images of the characters. However, in electronic digital processing, the 
 * character images are digitized by transforming them into data or 
 * algorithms that are stored in a computer system. The character shape data
 * is then used to display the images in the form of small dots on a 
 * presentation surface, for example, as picture elements (pels) on a display
 * screen or as dots on a piece of paper.
 *  <p/>
 * A digitized font contains not only the character shape information, but 
 * all of the information needed by an information processing system to 
 * format a character string and render the character shapes with a given 
 * presentation technology. Digitized fonts contain descriptive information
 * used to identify the specific font resource, metric information used to 
 * position the character shapes, and the shape information used to render 
 * the character shapes.
 *  <p/>
 * In theory, a font could be defined to be the univeral set of all the 
 * world's characters having the same design characteristics. In practice 
 * however, a digitized font resource contains a bounded set of characters 
 * having the same design characteristics. Further, multiple digitized font 
 * resources may exist having the same design characteristics, but each 
 * containing a different bounded set of characters. The set of characters 
 * contained in a font is determined by the font provider, depending on the 
 * language requirements for document processing.
 *  <p/>
 * Each graphic character in a font has a unique name, which is called the
 * graphic character identifier. Each graphic character in a document data
 * stream has a unique code, which is called a code point. A code page is a
 * table that associates graphic character identifiers with code points.
 * <p/>
 * When documents are formatted and presented, the graphic characters in the
 * document, which are usually encoded as code points of one or two bytes for
 * each graphic character, are associated with the character information in 
 * a font through the code page. For an example, presentation of the word 
 * "the" requires the character shapes for the graphic characters t, h, and 
 * e. Each code point in the code page maps to a character identifier, which 
 * points to the font information, including metrics and shape. One code page
 * is often sufficient to represent all the graphic characters in the 
 * document, although multiple code pages can be used.
 * <p/>
 * Fonts are either monospaced, in which the horizontal space occupied by 
 * different characters is the same, or typographic, that is, proportionally
 * spaced fonts, in which the character widths vary.
 * <p/>
 * In typographic fonts, the font designer has adjusted the distance from 
 * one character to another to improve the visual flow of text by eliminating
 * excess space. Adjusting space this way improves the readability and the 
 * appearance of a document.
 * <p/>
 * @author Pete Townsend <a href="mailto:pete@townsend.uk.com"/>
 */
public class Font {

	/**
	 * Static logging instance
	 */
	protected static final Log LOGGER =
		LogFactory.getLog("org.apache.fop.render.afp.fonts");

	/**
	 * The font family name
	 */
	private String _fontFamily = null;

	/**
	 * The code page for the font
	 */
	private String _codepage = null;

	/**
	 * The encoding that should be used with the code page
	 */
	private String _encoding = null;

	/**
	 * The collection of point sizes supported by this font
	 */
	private HashMap _pointSizes = null;

	/**
	 * The FontMetric information
	 */
	private HashMap _fontMetrics = null;

	/**
	 * Construct a font object.
	 * @param fontFamily the font family name
	 * @param codepage the codepage to be used
	 * @param encoding the encoding of the font
	 */
	public Font(String fontFamily, String codepage, String encoding) {

		_fontFamily = fontFamily;
		_codepage = codepage;
		_encoding = encoding;
		_pointSizes = new HashMap();
		_fontMetrics = new HashMap(4);

	}

	/**
	 * Add a font metrics object to the font.
	 * @param fontMetric the font metic to be added
	 */
	public void addFontMetric(FontMetrics fontMetric) {
		_fontMetrics.put(fontMetric.getFontMetricKey(), fontMetric);
	}

	/**
	 * Add a point size to this font.
	 * @param ps the point size to be added
	 */
	public void addPointSize(PointSize ps) {
		_pointSizes.put(String.valueOf(ps.getPointSize()), ps);
	}

	/**
	 * @return the font family name
	 */
	public String getFontFamily() {
		return _fontFamily;
	}

	/**
	 * @return the code page identifier
	 */
	public String getCodePage() {
		return _codepage;
	}

	/**
	 * Obtain the point size object for the specified value.
	 * @param pointSize a string representing the point size of the font
	 * @return the point size object                  
	 */
	public PointSize getPointSize(String pointSize) {

		if (pointSize == null) {
			throw new IllegalArgumentException("");
		}

		PointSize ps = null;
		String newPointSize = null;

		ps = (PointSize) _pointSizes.get(pointSize);

		if (ps == null) {

			LOGGER.warn(
				"Failed to locate a point size of "
					+ pointSize
					+ " in font family "
					+ _fontFamily);

			int i = Integer.parseInt(pointSize);
			int low = i - 1;
			int high = i + 1;

			// Try and locate a font within 10 points (higer or lower)
			for (int x = 0; i < 10; i++) {
				// try lower
				ps = getPointSize(String.valueOf(low));
				if (ps == null) {
					// try higher
					ps = getPointSize(String.valueOf(high));
				}
				if (ps != null) {
					// found a point size
					break;
				}
				low--;
				high++;
			}

			if (ps == null) {
				throw new FontRuntimeException(
					"Can't get an appropriate point size for pointSize :"
						+ pointSize);
			} else {
				LOGGER.warn(
					"Using point size "
						+ ps.getPointSize()
						+ " instead of "
						+ pointSize);
			}

		}

		return ps;
	}

	/**
	 * Obtain the character set binary data for font details passed.
	 * @param pointsize the point size
	 * @param weight the weight (i.e. normal/bold)
	 * @param style the style (i.e. normal/italics)
	 * @return the character set binary data
	 */
	public byte[] getCharacterSet(
		String pointsize,
		String weight,
		String style) {

		PointSize ps = (PointSize) _pointSizes.get(pointsize);
		if (ps == null) {
			throw new FontRuntimeException (
				"Can't get an appropriate point size for pointSize :"
					+ pointsize);
		}

		return ps.getCharacterSet(weight, style).getCharacterSet();

	}

	/**
	 * Returns the encoding that should be used for this font code page
	 * @return String
	 */
	public String getEncoding() {
		return _encoding;
	}

	/**
	 * @return the collection of font metrics
	 */
	public Collection getFontMetricCollection() {

		return _fontMetrics.values();

	}

	/**
	 * Obtain the font metric for the specified weight and style.
	 * @param weight the weight (i.e. normal/bold)
	 * @param style the style (i.e. normal/italics)
	 * @return the font metric 
	 */
	public FontMetric getFontMetric(String weight, String style) {

		return (FontMetric) _fontMetrics.get(weight + "-" + style);

	}

}
