/*
 * Copyright 1999-2004 The Apache Software Foundation.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/* $Id$ */

package org.apache.fop.render.afp.fonts;

import java.net.URL;
import java.util.Collection;
import java.util.HashMap;

import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.fop.render.afp.exceptions.FontRuntimeException;
import org.xml.sax.InputSource;
import org.xml.sax.XMLReader;

/**
 * The class provides a wrapper for all fonts installed which support
 * the specified dpi, the dpi will be used to determine the font files
 * that will be used to read the associated metrics.
 * <p/>
 * @author Pete Townsend <a href="mailto:pete@townsend.uk.com"/>
 */
public class FontCollection {
	
	/**
	 * Static logging instance
	 */
	protected static final Log LOGGER =
		LogFactory.getLog("org.apache.fop.render.afp.fonts");
	
	/**
	 * The AFP font configuration file containing font information.
	 */
	private static final String AFP_FONT_CONFIG = "afp-fonts.xml";

	/**
	 * The dpi resolution used for this font collection
	 */
	private int _dpi;

	/**
	 * Holds font information for a specific dpi
	 */
	private HashMap _fontCollection;

	/**
	 * Constructor for the font collection, requires the dpi resolution
	 * @param int the dpi resolution
	 */
	public FontCollection(int dpi) {

		_dpi = dpi;
		_fontCollection = new HashMap();
		load();

	}

	/**
	 * Method to add font objects to the font collection
	 * @param font the font to be added
	 */
	protected void addFont(Font font) {

		_fontCollection.put(font.getFontFamily(), font);

	}

	/**
	 * This method will obtain an Font object for the specified FontFamily.
	 * @param fontFamiliy the font family name
	 * @return afpFont the Font object
	 */
	public Font getFont(String fontFamily) {

		Font font = (Font) _fontCollection.get(fontFamily);

		if (font != null) {
			return font;
		} else {
			String msg =
				"FontCollection::getFont(String):: "
					+ fontFamily
					+ " not found in AFP font table";
			LOGGER.error(msg);
			throw new FontRuntimeException(msg);
		}

	}

	/**
	 * Method to return a collection of all the fonts for this dpi
	 * @return the collection of fonts
	 */
	public Collection getFonts() {

		return _fontCollection.values();

	}

	/**
	 * Method to load the afp font file, this reads the xml font config
	 * file using SAX and creates the associated fonts.
	 */
	private void load() {

		SAXParserFactory spf = SAXParserFactory.newInstance();
		spf.setNamespaceAware(true);
		spf.setValidating(true);

		try {

			FontHandler fontHandler = new FontHandler(this);

			// Create a JAXP SAXParser
			SAXParser saxParser = spf.newSAXParser();

			// Get the encapsulated SAX XMLReader
			XMLReader xmlReader = saxParser.getXMLReader();

			// Set the ContentHandler of the XMLReader
			xmlReader.setContentHandler(fontHandler);

			ClassLoader classLoader = Thread.currentThread().getContextClassLoader();
			if (classLoader == null) {
				classLoader = FontCollection.class.getClassLoader();
			}
			
			URL url = classLoader.getResource(AFP_FONT_CONFIG);
			InputSource is = new InputSource(url.openStream());
			
			// Parse the XML document
			xmlReader.parse(is);

		} catch (org.xml.sax.SAXException sex) {
			String msg = "SAX exception when parsing " + AFP_FONT_CONFIG;
			LOGGER.error(msg);
			throw new FontRuntimeException(msg,	sex);
		} catch (javax.xml.parsers.ParserConfigurationException pce) {
			String msg = "Parser Configuration exception when parsing " + AFP_FONT_CONFIG;
			LOGGER.error(msg);
			throw new FontRuntimeException(msg, pce);
		} catch (java.io.IOException iex) {
			String msg = "IO exception when parsing " + AFP_FONT_CONFIG;
			LOGGER.error(msg);
			throw new FontRuntimeException(msg, iex);
		}

	}

	/**
	 * Returns the dpi resolution of this font collection
	 * @return int
	 */
	public int getDPI() {
		return _dpi;
	}

}
