/*
 * Copyright 1999-2004 The Apache Software Foundation.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/* $Id$ */

package org.apache.fop.render.afp.fonts;

import java.io.UnsupportedEncodingException;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.StringTokenizer;

import org.apache.fop.render.afp.modca.AFPConstants;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;
/**
 *
 * This is the sax handler that will be responsible for parsing the
 * font details held in afp-fonts.xml file which should reside on the 
 * classpath. 
 * <p/>
 * @author Pete Townsend <a href="mailto:pete@townsend.uk.com"/>
 */
public class FontHandler extends DefaultHandler {

	/** Static element names */
	private static final String ELEMENT_ROOT = "installed-fonts";
	private static final String ELEMENT_AFPFONT = "afp-font";
	private static final String ELEMENT_POINTSIZE = "pointsize";
	private static final String ELEMENT_CHARACTERSET = "characterset";

	/** Static attribute names */
	private static final String ATTRIBUTE_PATH = "path";
	private static final String ATTRIBUTE_CODEPAGE = "codepage";
	private static final String ATTRIBUTE_ENCODING = "encoding";
	private static final String ATTRIBUTE_WEIGHT = "weight";
	private static final String ATTRIBUTE_STYLE = "style";
	private static final String ATTRIBUTE_FONT = "font";
	private static final String ATTRIBUTE_DPI = "dpi";
	private static final String ATTRIBUTE_ID = "id";

	/** the font collection object */
	private FontCollection _fontCollection = null;

	/** the dpi to be included */
	private int[] _dpi;

	/** the currnt font being processed */
	private Font _currentFont = null;

	/** the current point size being processed */
	private PointSize _currentPointSize = null;

	/** the map of font metrics */
	private HashMap _currentFontMetrics = null;

	/** the path attribute value */
	private String _path = null;

	/**
	 * Construct a new font handler for the specified collection.
	 * @param fontCollection the font collection to populate
	 */
	public FontHandler(FontCollection fontCollection) {

		_fontCollection = fontCollection;
		_currentFontMetrics = new HashMap(4);

	}

	/**
	 * @see org.xml.sax.ContentHandler#endElement(String, String, String)
	 */
	public void endElement(String arg0, String arg1, String arg2)
		throws SAXException {

		if (arg1.equals(ELEMENT_AFPFONT)) {

			boolean b = false;
			for (int i = 0; i < _dpi.length; i++) {
				if (_dpi[i] == _fontCollection.getDPI()) {
					b = true;
					break;
				}
			}

			if (b) {
				// Add the font metrics
				Collection c = _currentFontMetrics.values();
				Iterator i = c.iterator();
				while (i.hasNext()) {

					FontMetrics fontMetrics = (FontMetrics) i.next();
					_currentFont.addFontMetric(fontMetrics);

				}

				_fontCollection.addFont(_currentFont);

			}

			_currentFont = null;

		} else if (arg1.equals(ELEMENT_POINTSIZE)) {
			_currentFont.addPointSize(_currentPointSize);
			_currentPointSize = null;
		}
	}

	/**
	 * @see org.xml.sax.ContentHandler#startElement(String, String, String, Attributes)
	 */
	public void startElement(
		String uri,
		String name,
		String qName,
		Attributes atts)
		throws SAXException {

		String currentElement = null;
		if ("".equals(uri)) {
			currentElement = qName;
		} else {
			currentElement = name;
		}

		if (currentElement.equalsIgnoreCase(ELEMENT_ROOT)) {
			_path = atts.getValue(ATTRIBUTE_PATH);
			return;
		}

		if (currentElement.equalsIgnoreCase(ELEMENT_AFPFONT)) {

			//build the fonts
			String id = atts.getValue(ATTRIBUTE_ID);
			String encoding = atts.getValue(ATTRIBUTE_ENCODING);
			String dpi = atts.getValue(ATTRIBUTE_DPI);
			StringTokenizer st = new StringTokenizer(dpi, ",", false);
			_dpi = new int[st.countTokens()];
			int count = 0;
			while (st.hasMoreTokens()) {

				try {

					String s = st.nextToken();
					int i = Integer.parseInt(s);
					_dpi[count] = i;
					count++;

				} catch (NumberFormatException nfex) {

					throw new SAXException(nfex);

				}
			}

			// Reset font metrics map
			_currentFontMetrics.clear();

			String codepage = atts.getValue(ATTRIBUTE_CODEPAGE);
			_currentFont = new Font(id, codepage, encoding);
			return;

		}

		if (currentElement.equalsIgnoreCase(ELEMENT_POINTSIZE)) {

			String id = atts.getValue(ATTRIBUTE_ID);

			try {

				int size = Integer.parseInt(id);
				_currentPointSize = new PointSize(size);

			} catch (NumberFormatException nfex) {

				throw new SAXException(nfex);

			}

			return;

		}

		if (currentElement.equals(ELEMENT_CHARACTERSET)) {

			String weight = atts.getValue(ATTRIBUTE_WEIGHT);
			String style = atts.getValue(ATTRIBUTE_STYLE);
			String font = atts.getValue(ATTRIBUTE_FONT);

			try {

				Integer.parseInt(atts.getValue(ATTRIBUTE_DPI));

			} catch (NumberFormatException nfex) {

			}

			// See if the font metric already exist
			FontMetrics fontMetrics =
				(FontMetrics) _currentFontMetrics.get(
					FontMetrics.createFontMetricKey(weight, style));
			if (fontMetrics == null) {

				fontMetrics = new FontMetrics(weight, style);
				_currentFontMetrics.put(
					fontMetrics.getFontMetricKey(),
					fontMetrics);

			}

			CharacterSetMetric fontSpec =
				new CharacterSetMetric(
					String.valueOf(_currentFont.getCodePage()),
					_currentFont.getEncoding(),
					font,
					_path,
					_fontCollection.getDPI());

			fontMetrics.addFontSpec(_currentPointSize.getPointSize(), fontSpec);

			try {

				byte[] characterset =
					font.getBytes(AFPConstants.EBCIDIC_ENCODING);
				_currentPointSize.addCharacterSet(weight, style, characterset);
				return;

			} catch (UnsupportedEncodingException usee) {
				// throw a SAX exception if the encoding is not supported
				throw new SAXException(usee);
			}

		}

	}

}
