/*
 * Copyright 1999-2004 The Apache Software Foundation.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/* $Id$ */

package org.apache.fop.render.afp.fonts;

import java.util.HashMap;

import org.apache.fop.layout.FontMetric;

/**
 * An implementation of the {@link org.apache.fop.layout.FontMetric} interface
 * to provide the metric information of the AFP fonts.
 * <p/>
 * @author Pete Townsend <a href="mailto:pete@townsend.uk.com"/>
 */
public class FontMetrics implements FontMetric {
	
	/** The map of font metrics */
	private HashMap _fontMetrics;
	
	/** The first character */
	private int _firstCharacter = 0;
	
	/** The last character */
	private int _lastCharacter = 0;
	
	/** The weight (i.e. normal/bold) */
	private String _weight;
	
	/** The style (i.e. normal/italics) */
	private String _style;
	
	/**
	 * Constructs a FontMetrics object for the specified weight and style.
	 * @param weight the weight (i.e. normal/bold) 
	 * @param style the style (i.e. normal/italics)
	 */
	public FontMetrics(String weight, String style) {
		
		_weight = weight;
		_style = style;
		_fontMetrics = new HashMap();
		
	}
	
	/**
	 * Add the character set metrics for the specified point size.
	 * @param size the point size
	 * @param fontSpec the character set metrics
	 */
	public void addFontSpec(int size, CharacterSetMetric fontSpec) {
		_fontMetrics.put(String.valueOf(size), fontSpec);
	}

	/**
	 * @see org.apache.fop.layout.FontMetric#getAscender(int)
	 */
	public int getAscender(int size) {
		return getFontSpec(size).getAscender();
	}

	/**
	 * @see org.apache.fop.layout.FontMetric#getCapHeight(int)
	 */
	public int getCapHeight(int size) {
		return getFontSpec(size).getCapHeight();
	}

	/**
	 * @see org.apache.fop.layout.FontMetric#getDescender(int)
	 */
	public int getDescender(int size) {
		return getFontSpec(size).getDescender();
	}

	/**
	 * @see org.apache.fop.layout.FontMetric#getFirstChar()
	 */
	public int getFirstChar() {
		return _firstCharacter;
	}

	/**
	 * @see org.apache.fop.layout.FontMetric#getLastChar()
	 */
	public int getLastChar() {
		return _lastCharacter;
	}

	/**
	 * @see org.apache.fop.layout.FontMetric#getWidths(int)
	 */
	public int[] getWidths(int size) {
		return getFontSpec(size).getWidths();
	}

	/**
	 * @see org.apache.fop.layout.FontMetric#getXHeight(int)
	 */
	public int getXHeight(int size) {
		return getFontSpec(size).getXHeight();
	}

	/**
	 * @see org.apache.fop.layout.FontMetric#width(int, int)
	 */
	public int width(int i, int size) {
		return getFontSpec(size).width(i);
	}
	
	/**
	 * Getter for the font specification for the specified size.
	 * @param size the point size
	 * @return the character set metrics
	 */
	private CharacterSetMetric getFontSpec(int size) {
		return (CharacterSetMetric)_fontMetrics.get(String.valueOf(size/1000));
	}
	
	/**
     * @return the key calculated from the weight and style of this object
     */
    public String getFontMetricKey() {
        return createFontMetricKey(_weight, _style);
    }
    
    /**
     * Create a font metric key from given weight and style.
     * @param weight the weight (i.e. normal/bold)
     * @param style the style (i.e. normal/italics)
     * @return the font metric key
     */
    public static String createFontMetricKey(String weight, String style) {
    	return weight + "-" + style;
    }

	/**
	 * @return the style (i.e. normal/italics)
	 */
	public String getStyle() {
		return _style;
	}

	/**
	 * @return the weight (i.e. normal/bold)
	 */
	public String getWeight() {
		return _weight;
	}

}
