/*
 * Copyright 1999-2004 The Apache Software Foundation.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/* $Id$ */

package org.apache.fop.render.afp.modca;

import java.io.IOException;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.HashSet;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.fop.render.afp.tools.StringUtils;

/**
 * A data stream is a continuous ordered stream of data elements and 
 * objects conforming to a given format. Application programs can 
 * generate data streams destined for a presentation service, archive 
 * library, presentation device or another application program. The 
 * strategic presentation data stream architectures used is Mixed Object 
 * Document Content Architecture (MO:DCA). 
 * 
 * The MO:DCA architecture defines the data stream used by applications to 
 * describe documents and object envelopes for interchange with other 
 * applications and application services. Documents defined in the MO:DCA
 * format may be archived in a database, then later retrieved, viewed, 
 * annotated and printed in local or distributed systems environments. 
 * Presentation fidelity is accommodated by including resource objects 
 * in the documents that reference them.
 * 
 * @author Pete Townsend <a href="mailto:pete@townsend.uk.com"/>
 */
public class AFPDataStream {

	/**
	 * Static logging instance
	 */
	protected static final Log LOGGER =
		LogFactory.getLog("org.apache.fop.render.afp.modca");


	private static final String STATE_DOC_IN_PROGRESS = "DOC_IN_PROGRESS";
	private static final String STATE_PGE_IN_PROGRESS = "PGE_IN_PROGRESS";
	private static final String STATE_BNG_IN_PROGRESS = "BNG_IN_PROGRESS";

	/**
	 * Boolean completion indicator
	 */
	private boolean _complete = false;

	/**
	 * The application producing the AFP document
	 */
	private String _producer = null;

	/**
	 * The AFP document object
	 */
	private Document _document = null;

	/**
	 * The resolution dpi for this data stream
	 */
	private int _dpi;

	/**
	 * The AFPDataStream status
	 */
	private HashSet _stateSet = null;

	/**
	 * The constructor for the AFPDataStream requires an OutputStream
	 * as an argument, this is the output stream to which the AFP data
	 * will be written during the process of constructing the document.
	 * @param dpi the resolution dots per inch for the data stream
	 */
	public AFPDataStream(int dpi) {

		_dpi = dpi;
		_stateSet = new HashSet();

	}

	/**
	 * The document is started by invoking this method which creates 
	 * an instance of the AFP Document object and registers the start
	 * with a validation map which ensures that methods are not invoked
	 * out of the correct sequence.
	 */
	public void startDocument() {

		if (_complete) {
			String msg = "Invalid state - data stream completed";
			LOGGER.warn("startDocument():: " + msg);
			throw new IllegalStateException(msg);
		}

		if (_stateSet.contains(STATE_DOC_IN_PROGRESS)) {
			String msg = "Invalid State - document already in progress";
			LOGGER.warn("startDocument():: " + msg);
			throw new IllegalStateException(msg);
		}

		_stateSet.add(STATE_DOC_IN_PROGRESS);

		_document = new Document();

	}

	/**
	 * The document is ended by invoking this method which creates 
	 * an instance of the AFP Document object and registers the start
	 * with a validation map which ensures that methods are not invoked
	 * out of the correct sequence.
	 */
	public void endDocument() {

		if (_complete) {
			String msg = "Invalid state - data stream completed";
			LOGGER.warn("endDocument():: " + msg);
			throw new IllegalStateException(msg);
		}

		if (!_stateSet.contains(STATE_DOC_IN_PROGRESS)) {
			String msg = "Invalid State - document not in progress";
			LOGGER.warn("endDocument():: " + msg);
			throw new IllegalStateException(msg);
		}

	}

	/**
	 * Helper method which will construct a page. The page object is held
	 * in memory until processing has finished on the current page (marked
	 * by calling the endPage() method.
	 * @param pageWidth the width of the page
	 * @param pageHeight the height of the page
	 */
	public void startPage(int pageWidth, int pageHeight) {

		String pageName =
			"PGN"
				+ StringUtils.lpad(
					String.valueOf(_document.getPageCount() + 1),
					'0',
					5);

		Page page = new Page(pageName, pageWidth, pageHeight);

		_document.addPage(page);

	}

	/**
	 * Helper method to mark the end of the page. 
	 */
	public void endPage() {

		_document.getCurrentPage().endPage();

	}

	/**
	 * Helper method to create a map coded font object on the current page, 
	 * this method delegates the construction of the map coded font object 
	 * to the active environment group on the current page.
	 * @param fontReference the font number used as the resource identifier
	 * @param fontFamily the font family identifies the font name
	 * @param weight the weight of the font (e.g. normal, bold)
	 * @param style the style of the font (e.g. normal, italics)
	 * @param size the point size of the font
	 */
	public void createFont(
		byte fontReference,
		String fontFamily,
		String weight,
		String style,
		int size) {

		Page page = _document.getCurrentPage();
		page.createFont(fontReference, fontFamily, weight, style, size, _dpi);

	}

	/**
	 * Helper method to create text on the current page, this method delegates
	 * to the current presentation text object in order to construct the text.
	 * @param fontNumber the font number used as the resource identifier
	 * @param x the x coordinate of the text
	 * @param y the y coordinate of the text
	 * @param data the text data to create
	 */
	public void createText(int fontNumber, int x, int y, byte[] data) {

		_document.getCurrentPage().createText(fontNumber, x, y, data);

	}

	/**
	 * Method to create a line on the current page.
	 * @param x1 the first x coordinate of the line
	 * @param y1 the first y coordinate of the line
	 * @param x2 the second x coordinate of the line
	 * @param y2 the second y coordinate of the line
	 * @param thickness the thickness of the line
	 */
	public void createLine(int x1, int y1, int x2, int y2, int thickness) {

		_document.getCurrentPage().createLine(x1, y1, x2, y2, thickness);

	}

	/**
	 * Sets the application producing the AFP.
	 * @param producer the application producing the AFP datastream
	 */
	public void setProducer(String producer) {
		_producer = producer;
	}

	/**
	 * This method will create shading on the page using the specified 
	 * ooordinates (the shading contrast is controlled via the red, green
	 * blue parameters, by converting this to grey scale).
	 * @param x the x coordinate of the shading
	 * @param y the y coordinate of the shading
	 * @param w the width of the shaded area
	 * @param h the height of the shaded area
	 * @param red the red value
	 * @param green the green value
	 * @param blue the blue value
	 */
	public void createShading(
		int x,
		int y,
		int w,
		int h,
		int red,
		int green,
		int blue) {

		_document.getCurrentPage().createShading(x, y, w, h, red, green, blue);

	}

	/**
	 * Write the document data to the output stream, this will invoke
	 * getDataStream on the document, which is then invoked recursively
	 * on child AFPObject's to construct the full data stream.
	 * @param outputstream the output stream to which data is written
	 * @throws java.io.IOException.
	 */
	public void write(OutputStream outputstream) throws IOException {

		outputstream.write(_document.getDataStream());

	}

	/**
	 * Helper method which allows creation of the MPO 
	 * object, via the AEG. And the IPO via the Page.
	 * (See actual object for descriptions.)
	 * @param name the name of the static overlay
	 * @param orientation the orientation for the overlay (0, 90, 180, 270)
	 */
	public void createStaticOverlay(String name, int orientation) {

		Page page = _document.getCurrentPage();
		page.createIncludePageOverlay(name, orientation);
		ActiveEnvironmentGroup aeg = page.getActiveEnvironmentGroup();
		aeg.createOverlay(name);

	}

	/**
	 * Helper method which allows creation of the IPS 
	 * object, via the Page Object.
	 *  (See actual object for descriptions.)
	 * @param name the name of the dynamic overlay
	 * @param xCoor the x coordinate for where to place the overlay
	 * @param yCoor the y coordinate for where to place the overlay
	 */
	public void createDynamicOverlay(String name, int xCoor, int yCoor) {

		Page page = _document.getCurrentPage();
		page.createIncludePageSegment(name, xCoor, yCoor);

	}

	/**
	 * Helper which allows TLE's to be added to the page 
	 * @param attributes the list of names and values
	 */
	public void createPageTLE(ArrayList attributes) {

		Page page = _document.getCurrentPage();

		for (int i = 0; i < attributes.size(); i++) {

			String name = (String) attributes.get(i);
			i++;
			String value = (String) attributes.get(i);
			page.createTLE(name, value);

		}

	}
	
	/**
	 * Helper which allows TLE's to be added to the page group 
	 * @param attributes - this is the list of names and values, used for the indexing
	 */
	public void createPageGroupTLE(ArrayList attributes) {

		PageGroup pg = _document.get_currentPageGroup();
		LOGGER.debug("Number of TLE's to be added to the page group: "
				+ (attributes.size() / 2));

		for (int i = 0; i < attributes.size(); i++) {

			String name = (String) attributes.get(i);
			i++;
			String value = (String) attributes.get(i);
			pg.createPGTLE(name, value);

		}

	}
	/**
	 * Does as it says, creates a PageGroup object
	 * @param pageGroupName - the name assigned to the page group
	 */

	public void createPageGroup(String pageGroupName) {

		if (pageGroupName.length() != 8) {
			throw new IllegalArgumentException(
				"PageGroup name must be 8 characters long: " + pageGroupName);
		}

		PageGroup pg = new PageGroup(pageGroupName);
		_document.addPageGroup(pg);
	}
	/**
	 * Ends the PageGroup.
	 * @param pageGroupName - name value of afp:end-page-group xml tag.
	 * @throws IndexingException - this is thrown if you are trying to end a pg 
	 * that is not the current one.
	 */
	public void endPageGroup(String pageGroupName) throws IndexingException {

		if (pageGroupName.length() != 8) {
			throw new IllegalArgumentException(
				"PageGroup name must be 8 characters long: " + pageGroupName);
		}
		_document.endPageGroup(pageGroupName);

	}

}