/*
 * Copyright 1999-2004 The Apache Software Foundation.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/* $Id$ */

package org.apache.fop.render.afp.modca;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.util.ArrayList;
import java.util.Iterator;

/**
 * Pages contain the data objects that comprise a presentation document. Each 
 * page has a set of data objects associated with it. Each page within a
 * document is independent from any other page, and each must establish its 
 * own environment parameters.
 * 
 * The page is the level in the document component hierarchy that is used
 * for printing or displaying a document's content. The data objects contained 
 * in the page envelope in the data stream are presented when the page is 
 * presented. Each data object has layout information associated with it that 
 * directs the placement and orientation of the data on the page. In addition, 
 * each page contains layout information that specifies the measurement units, 
 * page width, and page depth.
 * 
 * A page is initiated by a begin page structured field and terminated by an
 * end page structured field. Structured fields that define objects and
 * active environment groups or that specify attributes of the page may be 
 * encountered in page state.
 * 
 * @author Pete Townsend <a href="mailto:pete@townsend.uk.com"/>
 */
public abstract class AbstractPage extends AFPObject {

	/**
	 * The name of the page object
	 */
	protected String _name = null;

	/**
	 * The name of the page object as EBCIDIC bytes
	 */
	protected byte[] _nameBytes;

	/**
	 * The active environment group for the page
	 */
	private ActiveEnvironmentGroup _activeEnvironmentGroup = null;

	/**
	 * The presentation text object, we only have one per page
	 */
	private PresentationTextObject _presentationTextObject = null;

	/**
	 * The list of image objects within the page
	 */
	private ArrayList _imageObjects = new ArrayList();

	/**
	 * The list of the include page objects
	 */
	private ArrayList _overlays = new ArrayList();

	/**
	 * Portrait or Landscape boolean indicator (deafault portrait)
	 */
	private boolean _portrait = true;
	
	/**
	 * The list of the include page segments
	 */
	private ArrayList _segments = new ArrayList();

	/**
	 * The page width
	 */
	private int _width;

	/**
	 * The list of tle's
	 */
	private ArrayList _tles = new ArrayList();
	
	/**
	 * Portrait or Landscape boolean indicator (deafault portrait)
	 * The page height
	 */
	private int _height;

	/**
	 * Construct a new page object for the specified name argument,
	 * the page name should be an 8 character identifier.
	 * @param name The name of the page.
	 * @param width The width of the page.
	 * @param height The height of the page.
	 */
	public AbstractPage(String name, int width, int height) {

		if (name.length() != 8) {
			throw new IllegalArgumentException(
				"Page name must be 8 characters long: " + name);
		}

		_name = name;

		// Set the page as being landscape if width is greater than height
		if (width > height) {
			_portrait = false;
			_width = height;
			_height = width;
		} else {
			_width = width;
			_height = height;
		}
			
		try {

			_nameBytes = name.getBytes(AFPConstants.EBCIDIC_ENCODING);

		} catch (UnsupportedEncodingException usee) {

			_nameBytes = name.getBytes();
			LOGGER.warn("Constructor:: UnsupportedEncodingException translating the name "
					+ name);

		}
		
		/**
		 * Every page object must have an ActiveEnvironmentGroup
		 */
		_activeEnvironmentGroup = new ActiveEnvironmentGroup(_width, _height);

		/**
		 *  We have a presentation text object per page
		 */
		_presentationTextObject = new PresentationTextObject();

	}

	/**
	 * Helper method to create a map coded font object on the current page, 
	 * this method delegates the construction of the map coded font object 
	 * to the active environment group on the page.
	 * @param fontReference The font number used as the resource identifier.
	 * @param fontFamily The font family identifies the font name.
	 * @param weight The font weight (e.g. normal, bold).
	 * @param style The font style (e.g. normal, italics).
	 * @param size The point size of the font.
	 * @param dpi The resolution to be used (dots per inch).
	 */
	public void createFont(
		byte fontReference,
		String fontFamily,
		String weight,
		String style,
		int size,
		int dpi) {

		_activeEnvironmentGroup.createFont(
			fontReference,
			fontFamily,
			weight,
			style,
			size,
			0,
			dpi);

	}

	/**
	 * Helper method to create a line on the current page, this method delegates
	 * to the presentation text object in order to construct the line.
	 * @param x1 The first x coordinate of the line.
	 * @param y1 The first y coordinate of the line.
	 * @param x2 The second x coordinate of the line.
	 * @param y2 The second y coordinate of the line.
	 * @param thickness The thickness of the line.
	 */
	public void createLine(int x1, int y1, int x2, int y2, int thickness) {

		if (!_portrait) {

			int a = x1;
			int b = y1;
			int c = x2;
			int d = y2;

			x1 = b;
			y1 = a;
			x2 = d;
			y2 = c;

		}

		_presentationTextObject.createLineData(x1, y1, x2, y2, thickness);

	}

	/**
	 * Helper method to create text on the current page, this method delegates
	 * to the presentation text object in order to construct the text.
	 * @param fontNumber The font number used as the resource identifier.
	 * @param x The x coordinate of the text data.
	 * @param y The y coordinate of the text data.
	 * @param data The text data to create.
	 */
	public void createText(int fontNumber, int x, int y, byte[] data) {

		int orientation = 0;

		if (!_portrait) {

			orientation = 270;

		}

		_presentationTextObject.createTextData(
			fontNumber,
			x,
			y,
			orientation,
			data);

	}

	/**
	 * Helper method to mark the end of the page. This should
	 * end the control sequence on the current presenation 
	 * text object. 
	 */
	public void endPage() {

		_presentationTextObject.endControlSequence();

	}

	/**
	 * This method will create shading on the page using the specified 
	 * ooordinates (the shading contrast is controlled via the red, green
	 * blue parameters, by converting this to grey scale).
	 * @param x The x coordinate of the shading.
	 * @param y The y coordinate of the shading.
	 * @param w The width of the shaded area
	 * @param h The height of the shaded area
	 * @param red The red value.
	 * @param green The green value.
	 * @param blue The blue value.
	 */
	public void createShading(
		int x,
		int y,
		int w,
		int h,
		int red,
		int green,
		int blue) {

		if (!_portrait) {

			int x1 = x;
			int y1 = y;
			int x2 = w;
			int y2 = h;

			x = y1;
			y = x1;
			w = y2;
			h = x2;

		}

		// Convert the color to grey scale
		float shade = (float)((red * 0.3) + (green * 0.59) + (blue * 0.11));
		
		int greyscale = Math.round((shade / 255) * 16);
		
		ImageObjectIM io = new ImageObjectIM("IMG00001");

		ImageOutputControl ioc = new ImageOutputControl(0, 0);

		ImageInputDescriptor iid = new ImageInputDescriptor();

		ImageCellPosition icp = new ImageCellPosition(x, y);
		icp.setXFillSize(w);
		icp.setYFillSize(h);
		icp.setXSize(64);
		icp.setYSize(8);

		//defing this as a resource
		ImageRasterData ird = new ImageRasterData(ImageRasterPattern.getRasterData(greyscale));

		io.setImageOutputControl(ioc);
		io.setImageInputDescriptor(iid);
		io.setImageCellPosition(icp);
		io.setImageRastorData(ird);

		_imageObjects.add(io);

	}

	/**
	 * Accessor method to obtain the byte array AFP datastream for the page.
	 * @return byte[] data stream
	 */
	public byte[] getDataStream() throws IOException {

		ByteArrayOutputStream baos = new ByteArrayOutputStream();

		baos.write(getStart());

		baos.write(_activeEnvironmentGroup.getDataStream());

		Iterator i1 = _imageObjects.iterator();
		while (i1.hasNext()) {
			ImageObjectIM io = (ImageObjectIM) i1.next();
			baos.write(io.getDataStream());
		}

		Iterator i2 = _overlays.iterator();
		while (i2.hasNext()) {
			IncludePageOverlay ipo = (IncludePageOverlay) i2.next();
			baos.write(ipo.getDataStream());
		}
		
		Iterator i3 = _segments.iterator();
		while(i3.hasNext()){
			IncludePageSegment ips= (IncludePageSegment)i3.next();
			baos.write(ips.getDataStream());
		}
		Iterator i4 = _tles.iterator();
		while(i4.hasNext()){
			TagLogicalElement tle= (TagLogicalElement)i4.next();
			baos.write(tle.getDataStream());
		}

		baos.write(_presentationTextObject.getDataStream());

		baos.write(getEnd());

		return baos.toByteArray();

	}

	/**
	 * Deals with the creation of the IPO object(see: IncludePageOverlay), 
	 * and the adding of it to the map.
	 * @param name - the name of the overlay
	 * @param orientation - the orientation required for the overlay
	 */
	public void createIncludePageOverlay(String name, int orientation){
		
		IncludePageOverlay ipo = new IncludePageOverlay(name,orientation);
		_overlays.add(ipo);
	}
	/**
	 * Deals with the creation of the TLE object(see: TagLogicalElement), 
	 * and the adding of it to the map.
	 * @param name - the name of the key to index on
	 * @param value - the value of the key
	 */
	public void createTLE(String name, String value){
		
		TagLogicalElement tle = new TagLogicalElement(name,value);
		_tles.add(tle);
		
	}
	/**
	 * Deals with the creation of the IPS (see: IncludePageSegment) object, 
	 * and the adding of it to the map.
	 * @param name - the name of the page segment to use
	 * @param xCoor - the x cooridinate for the page segment to be positioned at.
	 * @param yCoor - the y cooridinate for the page segment to be positioned at.
	 */
	public void createIncludePageSegment(String name, 
		int xCoor, 
		int yCoor){
		
		IncludePageSegment ips = new IncludePageSegment(name, xCoor,yCoor);
		_segments.add(ips);
		
	}

	/**
	 * Helper method to return the start of the page as a byte array.
	 * @return byte[] data stream
	 */
	protected abstract byte[] getStart();

	/**
	 * Helper method to return the end of the page as a byte array.
	 * @return byte[] data stream
	 */
	protected abstract byte[] getEnd();
	
	/**
	 * @return
	 */
	public ActiveEnvironmentGroup getActiveEnvironmentGroup() {
		return _activeEnvironmentGroup;
	}

}