/*
 * Copyright 1999-2004 The Apache Software Foundation.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/* $Id$ */

package org.apache.fop.render.afp.modca;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.util.ArrayList;
import java.util.Iterator;

/**
 * An Active Environment Group (AEG) is associated with each page, 
 * and is contained in the page's begin-end envelope in the data stream. 
 * The active environment group contains layout and formatting information 
 * that defines the measurement units and size of the page, and may contain 
 * resource information.
 * 
 * Any objects that are required for page presentation and that are to be 
 * treated as resource objects must be mapped with a map structured field
 * in the AEG. The scope of an active environment group is the scope of its
 * containing page or overlay. 
 * 
 * @author Pete Townsend <a href="mailto:pete@townsend.uk.com"/>
 */
public final class ActiveEnvironmentGroup extends AFPObject {

	/**
	 * Default name for the active environment group
	 */
	private static final String DEFAULT_NAME = "AEG00001";

	/**
	 * The actual name of the active enviornment group
	 */
	private String _name = null;

	/**
	 * The name of the page object as EBCIDIC bytes
	 */
	private byte[] _nameBytes;

	/**
	 * The collection of MapCodedFont objects
	 */
	private ArrayList _mapCodedFonts = new ArrayList();

	/**
	 * The PageDescriptor for the active environment group
	 */
	private PageDescriptor _pageDescriptor = null;

	/**
	 * The PresentationTextDescriptor for the active environment group
	 */
	private PresentationTextDescriptor _presentationTextDataDescriptor = null;

	/**
	 * The collection of MapPageOverlay objects
	 */
	private ArrayList _mapPageOverlays = new ArrayList();

	/**	
	 * Default constructor for the ActiveEnvironmentGroup.
	 * @param width the page width
	 * @param height the page height
	 */
	public ActiveEnvironmentGroup(int width, int height) {

		this(DEFAULT_NAME, width, height);

	}

	/**
	 * Constructor for the ActiveEnvironmentGroup, this takes a 
	 * name parameter which must be 8 characters long.
	 * @param name the active environment group name
	 * @param width the page width
	 * @param height the page height
	 */
	public ActiveEnvironmentGroup(String name, int width, int height) {

		try {

			_nameBytes = name.getBytes(AFPConstants.EBCIDIC_ENCODING);

		} catch (UnsupportedEncodingException usee) {

			_nameBytes = name.getBytes();
			LOGGER.warn(
				"Constructor:: UnsupportedEncodingException translating the name "
					+ name);

		}

		// Create PageDescriptor
		_pageDescriptor = new PageDescriptor(width, height);

		// Create PresentationTextDataDescriptor
		_presentationTextDataDescriptor =
			new PresentationTextDescriptor(width, height);

	}

	/**
	 * Accessor method to obtain the PageDescriptor object of the
	 * active environment group.
	 * @return the page descriptor object
	 */
	public PageDescriptor getPageDescriptor() {

		return _pageDescriptor;

	}

	/**
	 * Accessor method to obtain the PresentationTextDataDescriptor object of 
	 * the active environment group.
	 * @return the presentation text descriptor
	 */
	public PresentationTextDescriptor getPresentationTextDataDescriptor() {

		return _presentationTextDataDescriptor;

	}

	/**
	 * Accessor method to obtain the byte array AFP datastream for
	 * the active environment group.
	 * @return byte[] data stream
	 */
	public byte[] getDataStream() throws IOException {

		ByteArrayOutputStream baos = new ByteArrayOutputStream();

		baos.write(getStart());

		Iterator iterator = _mapCodedFonts.iterator();
		while (iterator.hasNext()) {
			MapCodedFont mcp = (MapCodedFont) iterator.next();
			baos.write(mcp.getDataStream());
		}
		Iterator iterator2 = _mapPageOverlays.iterator();
		while (iterator2.hasNext()) {
			MapPageOverlay mpo = (MapPageOverlay) iterator2.next();
			baos.write(mpo.getDataStream());
		}

		baos.write(_pageDescriptor.getDataStream());

		baos.write(_presentationTextDataDescriptor.getDataStream());

		baos.write(getEnd());

		return baos.toByteArray();

	}

	/**
	 * Helper method to return the start of the active environment group.
	 * @return byte[] the data stream
	 */
	private byte[] getStart() {

		byte[] data = new byte[17];

		data[0] = 0x5A; // Structured field identifier
		data[1] = 0x00; // Length byte 1
		data[2] = 0x10; // Length byte 2
		data[3] = (byte) 0xD3; // Structured field id byte 1
		data[4] = (byte) 0xA8; // Structured field id byte 2
		data[5] = (byte) 0xC9; // Structured field id byte 3
		data[6] = 0x00; // Flags
		data[7] = 0x00; // Reserved
		data[8] = 0x00; // Reserved

		for (int i = 0; i < _nameBytes.length; i++) {

			data[9 + i] = _nameBytes[i];

		}

		return data;

	}

	/**
	 * Helper method to return the end of the active environment.
	 * @return byte[] the data stream
	 */
	private byte[] getEnd() {

		byte[] data = new byte[17];

		data[0] = 0x5A; // Structured field identifier
		data[1] = 0x00; // Length byte 1
		data[2] = 0x10; // Length byte 2
		data[3] = (byte) 0xD3; // Structured field id byte 1
		data[4] = (byte) 0xA9; // Structured field id byte 2
		data[5] = (byte) 0xC9; // Structured field id byte 3
		data[6] = 0x00; // Flags
		data[7] = 0x00; // Reserved
		data[8] = 0x00; // Reserved

		for (int i = 0; i < _nameBytes.length; i++) {

			data[9 + i] = _nameBytes[i];

		}

		return data;

	}

	/**
	 * Method to create a map coded font object
	 * @param fontReference the font number used as the resource identifier
	 * @param fontFamily the font family identifies the font name
	 * @param weight the font weight (e.g. normal, bold)
	 * @param style the font style (e.g. normal, italics)
	 * @param size the point size of the font
	 * @param orientation the orientation of the font (e.g. 0, 90, 180, 270)
	 * @param dpi the resolution to be used (dots per inch)
	 */
	public void createFont(
		byte fontReference,
		String fontFamily,
		String weight,
		String style,
		int size,
		int orientation,
		int dpi) {

		MapCodedFont mcf = getCurrentMapCodedFont();

		if (mcf == null) {
			mcf = new MapCodedFont();
			_mapCodedFonts.add(mcf);
		}

		try {

			mcf.addFont(
				fontReference,
				fontFamily,
				size,
				weight,
				style,
				orientation,
				dpi);

		} catch (MaximumSizeExceededException msee) {

			mcf = new MapCodedFont();
			_mapCodedFonts.add(mcf);

			try {

				mcf.addFont(
					fontReference,
					fontFamily,
					size,
					weight,
					style,
					orientation,
					dpi);

			} catch (MaximumSizeExceededException ex) {

				// Should never happen (but log just in case)
				LOGGER.error("createFont():: resulted in a MaximumSizeExceededException");

			}

		}

	}

	/**
	 * Actually creates the MPO object.
	 * Also creates the supporting object (an IPO)
	 * @param name the name of the overlay to be used
	 */
	public void createOverlay(String name) {

		MapPageOverlay mpo = getCurrentMapPageOverlay();

		if (mpo == null) {
			mpo = new MapPageOverlay();
			_mapPageOverlays.add(mpo);
		}

		try {

			mpo.addOverlay(name);

		} catch (MaximumSizeExceededException msee) {
			mpo = new MapPageOverlay();
			_mapPageOverlays.add(mpo);
			try {
				mpo.addOverlay(name);
			} catch (MaximumSizeExceededException ex) {
				// Should never happen (but log just in case)
				LOGGER.error("createOverlay():: resulted in a MaximumSizeExceededException");
			}
		}
	}

	/**
	 * Getter method for the most recent MapCodedFont added to the 
	 * Active Environment Group (returns null if no MapCodedFonts exist)
	 * @return the most recent Map Coded Font.
	 */
	private MapCodedFont getCurrentMapCodedFont() {

		int size = _mapCodedFonts.size();
		if (size > 0) {
			return (MapCodedFont) _mapCodedFonts.get(_mapCodedFonts.size() - 1);
		} else {
			return null;
		}

	}

	/**
	 * Getter method for the most recent MapPageOverlay added to the 
	 * Active Environment Group (returns null if no MapPageOverlay exist)
	 * @return the most recent Map Coded Font
	 */
	private MapPageOverlay getCurrentMapPageOverlay() {

		int size = _mapPageOverlays.size();
		if (size > 0) {
			return (MapPageOverlay) _mapPageOverlays.get(
				_mapPageOverlays.size() - 1);
		} else {
			return null;
		}

	}

}