/*
 * Copyright 1999-2004 The Apache Software Foundation.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/* $Id$ */

package org.apache.fop.render.afp.modca;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.util.ArrayList;
import java.util.Iterator;

/**
 * The document is the highest level of the MO:DCA data-stream document
 * component hierarchy. Documents can be made up of pages, and the pages,
 * which are at the intermediate level, can be made up of objects. Objects
 * are at the lowest level, and can be bar codes, graphics, images, and 
 * presentation text.
 * 
 * At each level of the hierarchy certain sets of MO:DCA data structures, 
 * called structured fields, are permissible. The document, pages and objects 
 * are bounded by structured fields that define their beginnings and their ends. 
 * These structured fields, called begin-end pairs, provide an envelope for the
 * data-stream components. This feature enables a processor of the data stream 
 * that is not fully compliant with the architecture to bypass those objects
 * that are beyond its scope, and to process the data stream to the best of its
 * abilities.
 * 
 * A presentation document is one that has been formatted and is intended for
 * presentation, usually on a printer or display device. A data stream containing
 * a presentation document should produce the same document content in the
 * same format on different printers or display devices dependent, however, 
 * on the capabilities of each of the printers or display devices. A presentation
 * document can reference resources that are to be included as part of the
 * document to be presented.
 * 
 * @author Pete Townsend <a href="mailto:pete@townsend.uk.com"/>
 */
public final class Document extends AFPObject {

	/**
	 * Ststic default name reference
	 */
	private final static String DEFAULT_NAME = "DOC00001";

	/**
	 * A list of the objects in the document
	 */
	private ArrayList _objects = new ArrayList();

	/**
	 * The current page object (always the last one created)
	 */
	private Page _currentPage = null;
	
	/**
	 * The current page group object (always the last one created)
	 */
	private PageGroup _currentPageGroup = null;
	
	/**
	 * The page group counter
	 */
	private int _pageGroupCounter = 0;

	/**
	 * The page counter
	 */
	private int _pageCounter = 0;

	/**
	 * The name of the document
	 */
	private String _name = null;
	
	/**
	 * The name of the presentation text object as EBCIDIC bytes
	 */
	private byte[] _nameBytes;

	/**
	 * Default constructor for the document object.
	 */
	public Document() {

		this(DEFAULT_NAME);

	}
	
	/**
	 * Constructor for the document object.
	 * @param name The name of the document
	 */
	public Document(String name) {

		if (name.length() != 8) {
			throw new IllegalArgumentException(
				"Document name must be 8 characters long: " + name);
		}

		_name = name;

		try {

			_nameBytes = name.getBytes(AFPConstants.EBCIDIC_ENCODING);

		} catch (UnsupportedEncodingException usee) {

			_nameBytes = name.getBytes();
			LOGGER.warn("Constructor:: UnsupportedEncodingException translating the name "
					+ name);

		}

	}

	/**
	 * Helper method to return a byte array of the the start of the document. 
	 * @return byte[] data stream
	 */
	private byte[] getStart() {

		byte[] data = new byte[17];

		data[0] = 0x5A; // Structured field identifier
		data[1] = 0x00; // Length byte 1
		data[2] = 0x10; // Length byte 2
		data[3] = (byte) 0xD3; // Structured field id byte 1
		data[4] = (byte) 0xA8; // Structured field id byte 2
		data[5] = (byte) 0xA8; // Structured field id byte 3
		data[6] = 0x00; // Flags
		data[7] = 0x00; // Reserved
		data[8] = 0x00; // Reserved

		for (int i = 0; i < _nameBytes.length; i++) {

			data[9 + i] = _nameBytes[i];

		}

		return data;

	}

	/**
	 * Helper method to return a byte array of the end of the document.
	 * @return byte[] data stream
	 */
	private byte[] getEnd() {

		byte[] data = new byte[17];

		data[0] = 0x5A; // Structured field identifier
		data[1] = 0x00; // Length byte 1
		data[2] = 0x10; // Length byte 2
		data[3] = (byte) 0xD3; // Structured field id byte 1
		data[4] = (byte) 0xA9; // Structured field id byte 2
		data[5] = (byte) 0xA8; // Structured field id byte 3
		data[6] = 0x00; // Flags
		data[7] = 0x00; // Reserved
		data[8] = 0x00; // Reserved

		for (int i = 0; i < _nameBytes.length; i++) {

			data[9 + i] = _nameBytes[i];

		}

		return data;

	}

	/**
	 * Adds a page to the document.
	 * 
	 * Changed by rjh, so if you have a PageGroup, you
	 * add the Page to the PageGroup, else you add it to the doc.
	 * @param page - the Page object
	 */
	public void addPage(Page page) {

		if (_pageGroupCounter > 1){
			LOGGER.debug("addPage():: Adding a page to page group " + _pageGroupCounter);
			get_currentPageGroup().addPage(page);	
		}else {
			_pageCounter++;
			_currentPage = page;
			_objects.add(page);
			LOGGER.debug("addPage():: Adding page number " + _pageCounter);

		}
	}
	/**
	 * Adds a pageGroup to the document
	 * @param pg - The PageGroup object
	 */
	public void addPageGroup(PageGroup pg) {

		_pageGroupCounter++;
		_currentPageGroup =pg;

		_objects.add(pg);
		LOGGER.debug("addPageGroup():: Adding page group number " + _pageGroupCounter);

	}

	/**
	 * Getter method for the current page object, return null
	 * if there are no pages on the document.
	 */
	public Page getCurrentPage() {

		return _currentPage;

	}

	/**
	 * Accessor method to obtain the byte array AFP datastream for the document.
	 * @return byte[] data stream
	 */
	public byte[] getDataStream() throws IOException {

		ByteArrayOutputStream baos = new ByteArrayOutputStream();

		baos.write(getStart());

		Iterator i = _objects.iterator();
		while (i.hasNext()) {

			AFPObject apo = (AFPObject) i.next();
			baos.write(apo.getDataStream());

		}

		baos.write(getEnd());

		return baos.toByteArray();

	}
	
	/**
	 * Getter method for the page count within the document.
	 * @return int The page count
	 */
	public int getPageCount() {
		
		return _pageCounter;
		
	}

	/**
	 * Returns the _currentPageGroup.
	 * @return Page
	 */
	public PageGroup get_currentPageGroup() {
		return _currentPageGroup;
	}

	/**
	 * Returns the _pageGroupCounter.
	 * @return int
	 */
	public int get_pageGroupCounter() {
		return _pageGroupCounter;
	}
	/**
	 * Ends the current page group, firstly doing validation to 
	 * ensure you are ending the correct pg.
	 * @param name - name of the current pg, to end.
	 * @throws IndexingException if the pg (name) requested to be closed 
	 * doesn't equal the current pg name.
	 */
	public void endPageGroup(String pageGroupName) throws IndexingException{
	
		//Firstly get the current pg, and its name, then check
		//they match
		String nameOfCurrentPG =get_currentPageGroup().get_name();
		//set the pg to null so you can't add to it once its been closed
		_currentPageGroup = null;
		_pageGroupCounter =0;
	
		if (!(nameOfCurrentPG.equals(pageGroupName))){
			throw new IndexingException("Trying to close a PageGroup that doesn't exist");
		}
		EndPageGroup epg = new EndPageGroup(pageGroupName);
		_objects.add(epg);
	}
}