/*
 * Copyright 1999-2004 The Apache Software Foundation.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/* $Id$ */

package org.apache.fop.render.afp.modca;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.UnsupportedEncodingException;

/**
 * An IM image data object specifies the contents of a raster image and 
 * its placement on a page, overlay, or page segment. An IM image can be 
 * either simple or complex. A simple image is composed of one or more Image
 * Raster Data (IRD) structured fields that define the raster pattern for the
 * entire image. A complex image is divided into regions called image cells. 
 * Each image cell is composed of one or more IRD structured fields that define
 * the raster pattern for the image cell, and one Image Cell Position (ICP) 
 * structured field that defines the position of the image cell relative to 
 * the origin of the entire image. Each ICP also specifies the size of the 
 * image cell and a fill rectangle into which the cell is replicated. 
 * <p/>
 * @author Pete Townsend <a href="mailto:pete@townsend.uk.com"/>
 */
public class ImageObjectIM extends AFPObject {

	/**
	 * The name of the image object
	 */
	private String _name;

	/**
	 * The name of the image object as EBCIDIC bytes
	 */
	private byte[] _nameBytes;

	/**
	 * The image output control
	 */
	private ImageOutputControl _imageOutputControl = null;

	/**
	 * The image input descriptor
	 */
	private ImageInputDescriptor _imageInputDescriptor = null;

	/**
	 * The image cell position
	 */
	private ImageCellPosition _imageCellPosition = null;

	/**
	 * The image rastor data
	 */
	private ImageRasterData _imageRastorData = null;

	/**
	 * Constructor for the image object with the specified name,
	 * the name must be a fixed length of eight characters.
	 * @param name	The name of the image.
	 */
	public ImageObjectIM(String name) {

		if (name.length() != 8) {
			String msg = "Image name must be 8 characters long " + name;
			LOGGER.error("Constructor:: " + msg);
			throw new IllegalArgumentException(msg);
		}

		_name = name;

		try {

			_nameBytes = name.getBytes(AFPConstants.EBCIDIC_ENCODING);

		} catch (UnsupportedEncodingException usee) {

			_nameBytes = name.getBytes();
			LOGGER.warn(
				"Constructor:: UnsupportedEncodingException translating the name "
					+ name);

		}

	}

	/**
	 * Helper method to return the start of the image object.
	 * @return byte[] The data stream.
	 */
	private byte[] getStart() {

		byte[] data = new byte[17];

		data[0] = 0x5A; // Structured field identifier
		data[1] = 0x00; // Length byte 1
		data[2] = 0x10; // Length byte 2
		data[3] = (byte) 0xD3; // Structured field id byte 1
		data[4] = (byte) 0xA8; // Structured field id byte 2
		data[5] = (byte) 0x7B; // Structured field id byte 3
		data[6] = 0x00; // Flags
		data[7] = 0x00; // Reserved
		data[8] = 0x00; // Reserved

		for (int i = 0; i < _nameBytes.length; i++) {

			data[9 + i] = _nameBytes[i];

		}

		return data;

	}

	/**
	 * Helper method to return the end of the image object.
	 * @return byte[] The data stream.
	 */
	private byte[] getEnd() {

		byte[] data = new byte[17];

		data[0] = 0x5A; // Structured field identifier
		data[1] = 0x00; // Length byte 1
		data[2] = 0x10; // Length byte 2
		data[3] = (byte) 0xD3; // Structured field id byte 1
		data[4] = (byte) 0xA9; // Structured field id byte 2
		data[5] = (byte) 0x7B; // Structured field id byte 3
		data[6] = 0x00; // Flags
		data[7] = 0x00; // Reserved
		data[8] = 0x00; // Reserved

		for (int i = 0; i < _nameBytes.length; i++) {

			data[9 + i] = _nameBytes[i];

		}

		return data;

	}

	/* (non-Javadoc)
	 * @see org.apache.fop.render.afp.modca.AFPObject#getDataStream()
	 */
	public byte[] getDataStream() throws IOException {

		ByteArrayOutputStream baos = new ByteArrayOutputStream();

		baos.write(getStart());

		if (_imageOutputControl != null) {
			baos.write(_imageOutputControl.getDataStream());
		}

		if (_imageInputDescriptor != null) {
			baos.write(_imageInputDescriptor.getDataStream());
		}

		if (_imageCellPosition != null) {
			baos.write(_imageCellPosition.getDataStream());
		}

		if (_imageRastorData != null) {
			baos.write(_imageRastorData.getDataStream());
		}

		baos.write(getEnd());

		return baos.toByteArray();

	}

	/**
	 * Sets the ImageOutputControl.
	 * @param imageOutputControl The imageOutputControl to set
	 */
	public void setImageOutputControl(ImageOutputControl imageOutputControl) {
		_imageOutputControl = imageOutputControl;
	}

	/**
	 * Sets the ImageCellPosition.
	 * @param imageCellPosition The imageCellPosition to set
	 */
	public void setImageCellPosition(ImageCellPosition imageCellPosition) {
		_imageCellPosition = imageCellPosition;
	}

	/**
	 * Sets the ImageInputDescriptor.
	 * @param imageInputDescriptor The imageInputDescriptor to set
	 */
	public void setImageInputDescriptor(ImageInputDescriptor imageInputDescriptor) {
		_imageInputDescriptor = imageInputDescriptor;
	}

	/**
	 * Sets the ImageRastorData.
	 * @param imageRastorData The imageRastorData to set
	 */
	public void setImageRastorData(ImageRasterData imageRastorData) {
		_imageRastorData = imageRastorData;
	}

}
