/*
 * Copyright 1999-2004 The Apache Software Foundation.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/* $Id$ */

package org.apache.fop.render.afp.modca;

import java.io.UnsupportedEncodingException;

import org.apache.fop.render.afp.tools.BinaryUtils;

/**
 *
 * The Include Page Overlay structured field references an overlay resource definition
 * that is to be positioned on the page. A page overlay can be referenced at any time
 * during the page state, but not during an object state. The overlay contains its own
 * active environment group definition.
 *  
 * Note: There is no need for the triplets, so I have ignored them.
 * 
 * A real example of where this will be used is for static
 * overlays, such as an address on the page.
 * 
 * Further note:  Although an IPO needs x, y coordinates in the MODCA standard
 * due to the way IPO's are implemented over on the mainframe (being static) 
 * I have 'hardwired' the x+y to 0, as they do infact not get used. So the only
 * thing that will be dynamic and controlled by the FO is the orientation.
 * 
 * @author Pete Townsend <a href="mailto:pete@townsend.uk.com"/>
 */
public class IncludePageOverlay extends AFPObject {

	/**
	 * The name of the ipo, as a String 
	 */
	private String _overlayName = null;

	/**
	 * The name of the ipo object as EBCIDIC bytes
	 */
	private byte[] _nameBytes;

	private int _xCoor = 0;
	private int _yCoor = 0;
	private int _orientation = 0;
	private final byte END_VAL = (byte) 0x00;
	private final byte ZERO_ORIENTATION = (byte) 0x00;
	private final byte NINETY_ORIENTATION = (byte) 0x2D;
	private final byte ONEEIGHTY_ORIENTATION = (byte) 0x5A;
	private final byte TWOSEVENTY_ORIENTATION = (byte) 0x87;
	private final byte ENDBIT = (byte) 0x00;

	public IncludePageOverlay(String overLayName, int orientation) {

		validateName(overLayName);
		if (overLayName.length() != 8) {
			throw new IllegalArgumentException(
				"Document name must be 8 characters long: " + overLayName);
		}

		_overlayName = overLayName;

		try {

			_nameBytes = overLayName.getBytes(AFPConstants.EBCIDIC_ENCODING);

		} catch (UnsupportedEncodingException usee) {

			_nameBytes = overLayName.getBytes();
			LOGGER.warn(
				"Constructor:: UnsupportedEncodingException translating the name "
					+ overLayName);

		}
		_orientation = orientation;
	}

	/* (non-Javadoc)
	 * @see org.apache.fop.render.afp.modca.AFPObject#getDataStream()
	 */
	public byte[] getDataStream() {

		byte[] data = new byte[25]; //(9 +16)

		data[0] = 0x5A;

		// Set the total record length
		byte[] rl1 = BinaryUtils.convert(24, 2); //Ignore first byte
		data[1] = rl1[0];
		data[2] = rl1[1];

		// Structured field ID for a IPO
		data[3] = (byte) 0xD3;
		data[4] = (byte) 0xAF;
		data[5] = (byte) 0xD8;

		data[6] = 0x00; // Reserved
		data[7] = 0x00; // Reserved
		data[8] = 0x00; // Reserved

		for (int i = 0; i < _nameBytes.length; i++) {

			data[9 + i] = _nameBytes[i];

		}

		byte[] r2 = BinaryUtils.convert(0, 3);
		data[17] = r2[0]; // x coordinate
		data[18] = r2[1];
		data[19] = r2[2];

		byte[] r3 = BinaryUtils.convert(0, 3);
		data[20] = r3[0]; // y coordinate
		data[21] = r3[1];
		data[22] = r3[2];

		//Values
		byte[] r4 = getOrientationValues();
		data[23] = r4[0]; // orientation
		data[24] = r4[1];

		return data;

	}
	/**
	 * This method will return the non-standard orientation value
	 * based on the required orientation. These are fixed and 
	 * won't change unless the MODCA spec does.
	 * @return byte [] -of 2 bytes with the hex values required
	 * to achieve the orientation specified.
	 */

	private byte[] getOrientationValues() {

		byte[] answer = new byte[2];

		if (_orientation == 0) {
			answer[0] = ZERO_ORIENTATION;
			answer[1] = END_VAL;
		}
		if (_orientation == 90) {
			answer[0] = NINETY_ORIENTATION;
			answer[1] = END_VAL;
		}
		if (_orientation == 180) {
			answer[0] = ONEEIGHTY_ORIENTATION;
			answer[1] = END_VAL;
		}
		if (_orientation == 270) {
			answer[0] = TWOSEVENTY_ORIENTATION;
			answer[1] = END_VAL;
		}

		return answer;
	}

	/**
	 * This makes sure that the standards for overlays are
	 * followed.
	 * @param name - the name of the overlay, as specified in the FO.
	 */
	public void validateName(String name) {

		//the name of the Overlay must start with a O1 (an ohh not a zero!)
		if (!(name.startsWith("O1", 0))) {
			throw new IllegalArgumentException("Name of the overlay must start with 01");
		}

	}
	/**
	 * This makes sure that the possible values for
	 * orientation are adhered to.
	 * @param orientation - the orientation value, as specified in the FO.
	 */
	private void validateOrientation(int orientation) {

		//Orientation must be 0,90,180 or 270
		if ((!(orientation == 0))
			|| (!(orientation == 90))
			|| (!(orientation == 180))
			|| (!(orientation == 270))) {
			throw new IllegalArgumentException(
				"Orientation of the IPO must be 0,90,180,270, it is: "
					+ orientation);
		}

	}

}
