/*
 * Copyright 1999-2004 The Apache Software Foundation.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/* $Id$ */

package org.apache.fop.render.afp.modca;

import java.io.UnsupportedEncodingException;
import java.util.ArrayList;

import org.apache.fop.render.afp.exceptions.FontRuntimeException;
import org.apache.fop.render.afp.fonts.Font;
import org.apache.fop.render.afp.fonts.FontCollection;
import org.apache.fop.render.afp.fonts.FontFactory;
import org.apache.fop.render.afp.tools.BinaryUtils;

/**
 * The Map Coded Font structured field maps a unique coded font resource 
 * local ID, which may be embedded one or more times within an object's 
 * data and descriptor, to the identifier of a coded font resource object. 
 * Additionally, the Map Coded Font structured field specifies a set of 
 * resource attributes for the coded font.
 * 
 * @author Pete Townsend <a href="mailto:pete@townsend.uk.com"/>
 */
public class MapCodedFont extends AFPObject {

	/**
	 * The collection of map coded fonts (maximum of 254)
	 */
	private ArrayList _fontlist = null;

	/**
	 * Constructor for the MapCodedFont 
	 */
	public MapCodedFont() {

		_fontlist = new ArrayList();

	}

	/**
	 * Accessor method to obtain the byte array AFP datastream for
	 * the map coded font.
	 * @return byte[] data stream
	 */
	public byte[] getDataStream() {

		int fontcount = _fontlist.size();
		int recordlength = fontcount * 34;

		byte[] data = new byte[recordlength + 9];

		data[0] = 0x5A;

		// Set the total record length
		byte[] rl1 = BinaryUtils.convert(recordlength + 8, 2);
		data[1] = rl1[0];
		data[2] = rl1[1];

		// Field identifier for a MapCodedFont
		data[3] = (byte) 0xD3;
		data[4] = (byte) 0xAB;
		data[5] = (byte) 0x8A;

		data[6] = 0x00; // Reserved
		data[7] = 0x00; // Reserved
		data[8] = 0x00; // Reserved

		int pos = 8;

		for (int i = 0; i < _fontlist.size(); i++) {

			FontDefinition fd = (FontDefinition) _fontlist.get(i);

			// Start of repeating groups (occurs 1 to 254) 
			data[++pos] = 0x00;
			data[++pos] = 0x22; // Length of 34

			// Font Character Set Name Reference 
			data[++pos] = 0x0C;
			data[++pos] = 0x02;
			data[++pos] = (byte) 0x86;
			data[++pos] = 0x00;

			int x = 0;

			for (x = 0; x < fd._characterset.length; x++) {
				data[++pos] = fd._characterset[x];
			}

			// Font Code Page Name Reference
			data[++pos] = 0x0C;
			data[++pos] = 0x02;
			data[++pos] = (byte) 0x85;
			data[++pos] = 0x00;

			for (x = 0; x < fd._codepage.length; x++) {
				data[++pos] = fd._codepage[x];
			}

			// Character Rotation 
			data[++pos] = 0x04;
			data[++pos] = 0x26;
			data[++pos] = fd._orientation;
			data[++pos] = 0x00;

			// Resource Local Identifier
			data[++pos] = 0x04;
			data[++pos] = 0x24;
			data[++pos] = 0x05;
			data[++pos] = fd._fontReferenceKey;

		}

		return data;

	}

	/**
	 * Add a font definition on the the map coded font object.
	 * @param fontReference The font number used as the resource identifier.
	 * @param font the font family name
	 * @param size the point size of the font
	 * @param dpi the resolution of the font in dots per inch
	 */
	public void addFont(byte fontReference, String font, int size, int dpi)
		throws MaximumSizeExceededException {
		addFont(fontReference, font, size, "normal", "normal", 0, dpi);
	}

	/**
	 * Add a font definition on the the map coded font object.
	 * @param fontReference the font number used as the resource identifier
	 * @param font the font family identifies the font name
	 * @param weight the weight of the font (e.g. normal, bold)
	 * @param style the style of the font (e.g. normal, italics)
	 * @param size the point size of the font
	 * @param dpi the resolution of the font in dots per inch
	 */
	public void addFont(
		byte fontReference,
		String name,
		int size,
		String weight,
		String style,
		int orientation,
		int dpi)
		throws MaximumSizeExceededException {

		FontDefinition fd = new FontDefinition();

		fd._fontReferenceKey = fontReference;

		switch (orientation) {
			case 90 :
				fd._orientation = 0x2D;
				break;
			case 180 :
				fd._orientation = 0x5A;
				break;
			case 270 :
				fd._orientation = (byte) 0x87;
				break;
			default :
				fd._orientation = 0x00;
				break;
		}

		try {

			FontCollection fontCollection =
				FontFactory.getInstance().getFontCollection(dpi);
			Font font = fontCollection.getFont(name);

			fd._codepage =
				font.getCodePage().getBytes(AFPConstants.EBCIDIC_ENCODING);

			if (fd._codepage.length != 8) {
				throw new IllegalArgumentException(
					"The code page "
						+ new String(fd._codepage, AFPConstants.EBCIDIC_ENCODING)
						+ " must have a fixed length of 8 characters.");
			}

			fd._characterset =
				font.getCharacterSet(String.valueOf(size), weight, style);

			if (fd._characterset.length != 8) {
				throw new IllegalArgumentException(
					"The character set "
						+ new String(
							fd._characterset,
							AFPConstants.EBCIDIC_ENCODING)
						+ " must have a fixed length of 8 characters.");
			}

			if (_fontlist.size() > 253) {

				// Throw an exception if the size is exceeded
				throw new MaximumSizeExceededException();

			} else {

				_fontlist.add(fd);

			}

		} catch (UnsupportedEncodingException ex) {

			throw new FontRuntimeException(
				"Failed to create font "
					+ name
					+ " due to a UnsupportedEncodingException",
				ex);

		}

	}

	/**
	 * Private utility class used as a container for font attributes
	 */
	private class FontDefinition {

		/**
		 * The code page of the font
		 */
		private byte[] _codepage;

		/**
		 * The character set of the font
		 */
		private byte[] _characterset;

		/**
		 * The font reference key
		 */
		private byte _fontReferenceKey;

		/**
		 * The orientation of the font
		 */
		private byte _orientation;

	}

}