/*
 * Copyright 1999-2004 The Apache Software Foundation.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/* $Id$ */

package org.apache.fop.render.afp.modca;


/**
 * Pages contain the data objects that comprise a presentation document. Each 
 * page has a set of data objects associated with it. Each page within a
 * document is independent from any other page, and each must establish its 
 * own environment parameters.
 * 
 * The page is the level in the document component hierarchy that is used
 * for printing or displaying a document's content. The data objects contained 
 * in the page envelope in the data stream are presented when the page is 
 * presented. Each data object has layout information associated with it that 
 * directs the placement and orientation of the data on the page. In addition, 
 * each page contains layout information that specifies the measurement units, 
 * page width, and page depth.
 * 
 * A page is initiated by a begin page structured field and terminated by an
 * end page structured field. Structured fields that define objects and
 * active environment groups or that specify attributes of the page may be 
 * encountered in page state.
 * 
 * @author Pete Townsend <a href="mailto:pete@townsend.uk.com"/>
 */
public class Page extends AbstractPage {

	/**
	 * Construct a new page object for the specified name argument,
	 * the page name should be an 8 character identifier.
	 * @param name The name of the page.
	 * @param width The width of the page.
	 * @param height The height of the page.
	 */
	public Page(String name, int width, int height) {
		super(name, width, height);
	}

	/**
	 * Helper method to return the start of the page as a byte array.
	 * @return byte[] data stream
	 */
	protected byte[] getStart() {

		byte[] data = new byte[17];

		data[0] = 0x5A; // Structured field identifier
		data[1] = 0x00; // Length byte 1
		data[2] = 0x10; // Length byte 2
		data[3] = (byte) 0xD3; // Structured field id byte 1
		data[4] = (byte) 0xA8; // Structured field id byte 2
		data[5] = (byte) 0xAF; // Structured field id byte 3
		data[6] = 0x00; // Flags
		data[7] = 0x00; // Reserved
		data[8] = 0x00; // Reserved

		for (int i = 0; i < _nameBytes.length; i++) {

			data[9 + i] = _nameBytes[i];

		}

		return data;

	}

	/**
	 * Helper method to return the end of the page as a byte array.
	 * @return byte[] data stream
	 */
	protected byte[] getEnd() {

		byte[] data = new byte[17];

		data[0] = 0x5A; // Structured field identifier
		data[1] = 0x00; // Length byte 1
		data[2] = 0x10; // Length byte 2
		data[3] = (byte) 0xD3; // Structured field id byte 1
		data[4] = (byte) 0xA9; // Structured field id byte 2
		data[5] = (byte) 0xAF; // Structured field id byte 3
		data[6] = 0x00; // Flags
		data[7] = 0x00; // Reserved
		data[8] = 0x00; // Reserved

		for (int i = 0; i < _nameBytes.length; i++) {

			data[9 + i] = _nameBytes[i];

		}

		return data;

	}

}