/*
 * Copyright 1999-2004 The Apache Software Foundation.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/* $Id$ */

package org.apache.fop.render.afp.modca;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.util.ArrayList;
import java.util.Iterator;

/**
 * The Presentation Text object is the data object used in document processing
 * environments for representing text which has been prepared for presentation. 
 * Text, as used here, means an ordered string of characters, such as graphic 
 * symbols, numbers, and letters, that are suitable for the specific purpose 
 * of representing coherent information. Text which has been prepared for 
 * presentation has been reduced to a primitive form through explicit 
 * specification of the characters and their placement in the presentation space.
 * Control sequences which designate specific control functions may be embedded 
 * within the text. These functions extend the primitive form by applying specific 
 * characteristics to the text when it is presented. The collection of the graphic
 * characters and control codes is called Presentation Text, and the object that 
 * contains the Presentation Text is called the PresentationText object.
 * 
 * @author Pete Townsend <a href="mailto:pete@townsend.uk.com"/>
 */
public class PresentationTextObject extends AFPObject {

	/**
	 * Default name for the presentation text object
	 */
	private static final String DEFAULT_NAME = "PTO00001";

	/**
	 * The name of the presentation text object
	 */
	private String _name;

	/**
	 * The name of the presentation text object as EBCIDIC bytes
	 */
	private byte[] _nameBytes;

	private PresentationTextData currentPresentationTextData = null;
	private ArrayList presentationTextData = new ArrayList();

	/**
	 * Default constructor for the PresentationTextObject
	 */
	public PresentationTextObject() {

		this(DEFAULT_NAME);

	}

	/**
	 * Construct a new PresentationTextObject for the specified name argument,
	 * the name should be an 8 character identifier.
	 */
	public PresentationTextObject(String name) {

		if (name.length() != 8) {
			throw new IllegalArgumentException("The name of the PresentationTextObject must be 8 characters");
		}

		_name = name;

		try {

			_nameBytes = name.getBytes(AFPConstants.EBCIDIC_ENCODING);

		} catch (UnsupportedEncodingException usee) {

			_nameBytes = name.getBytes();
			LOGGER.warn(
				"Constructor:: UnsupportedEncodingException translating the name "
					+ name);

		}

	}

	/**
	 * Create the prsentation text data for the byte array of data.
	 * @param fontNumber The font resource identifier.
	 * @param x The x coordinate for the text data.
	 * @param y The y coordinate for the text data.
	 * @param data The text data to be created.
	 */
	public void createTextData(int fontNumber, int x, int y, byte[] data) {

		// Use a default orientation of zero
		createTextData(fontNumber, x, y, 0, data);

	}

	/**
	 * Create the presentation text data for the byte array of data.
	 * @param fontNumber The font resource identifier.
	 * @param x The x coordinate for the text data.
	 * @param y The y coordinate for the text data.
	 * @param orientation The orientation of the text data.
	 * @param data The text data to be created.
	 */
	public void createTextData(
		int fontNumber,
		int x,
		int y,
		int orientation,
		byte[] data) {

		if (currentPresentationTextData == null) {
			startPresentationTextData();
		}

		try {

			currentPresentationTextData.createTextData(
				fontNumber,
				x,
				y,
				orientation,
				data);

		} catch (MaximumSizeExceededException msee) {

			endPresentationTextData();
			createTextData(fontNumber, x, y, orientation, data);

		}

	}

	/**
	 * Create line data (i.e. drawing of lines) using the starting and ending
	 * coordinates, thickness and colour arguments.
	 * @param x1 The first x coordinate of the line.
	 * @param y1 The first y coordinate of the line.
	 * @param x2 The second x coordinate of the line.
	 * @param y2 The second y coordinate of the line.
	 * @param thickness The thickness of the line.
	 */
	public void createLineData(int x1, int y1, int x2, int y2, int thickness) {

		if (currentPresentationTextData == null) {
			startPresentationTextData();
		}

		try {

			currentPresentationTextData.createLineData(
				x1,
				y1,
				x2,
				y2,
				thickness);

		} catch (MaximumSizeExceededException msee) {

			endPresentationTextData();
			createLineData(x1, y1, x2, y2, thickness);

		}

	}

	/**
	 * Helper method to mark the start of the presnetation text data
	 */
	private void startPresentationTextData() {

		if (presentationTextData.size() == 0) {
			currentPresentationTextData = new PresentationTextData(true);
		} else {
			currentPresentationTextData = new PresentationTextData();
		}

		presentationTextData.add(currentPresentationTextData);

	}

	/**
	 * Helper method to mark the end of the presnetation text data
	 */
	private void endPresentationTextData() {

		currentPresentationTextData = null;

	}

	/**
	 * Accessor method to obtain the byte array AFP datastream for the 
	 * PresentationTextObject.
	 * @return byte[] data stream
	 */
	public byte[] getDataStream() throws IOException {

		ByteArrayOutputStream baos = new ByteArrayOutputStream();

		baos.write(getStart());

		Iterator iterator = presentationTextData.iterator();
		while (iterator.hasNext()) {
			PresentationTextData ptd = (PresentationTextData) iterator.next();
			baos.write(ptd.getDataStream());
		}

		baos.write(getEnd());

		return baos.toByteArray();

	}

	public String getName() {

		return _name;

	}

	/**
	 * Helper method to return the start of the presenation text object.
	 * @return byte[] data stream
	 */
	private byte[] getStart() {

		byte[] data = new byte[17];

		data[0] = 0x5A; // Structured field identifier
		data[1] = 0x00; // Length byte 1
		data[2] = 0x10; // Length byte 2
		data[3] = (byte) 0xD3; // Structured field id byte 1
		data[4] = (byte) 0xA8; // Structured field id byte 2
		data[5] = (byte) 0x9B; // Structured field id byte 3
		data[6] = 0x00; // Flags
		data[7] = 0x00; // Reserved
		data[8] = 0x00; // Reserved

		for (int i = 0; i < _nameBytes.length; i++) {

			data[9 + i] = _nameBytes[i];

		}

		return data;

	}

	/**
	 * Helper method to return the end of the presenation text object.
	 * @return byte[] data stream
	 */
	private byte[] getEnd() {

		byte[] data = new byte[17];

		data[0] = 0x5A; // Structured field identifier
		data[1] = 0x00; // Length byte 1
		data[2] = 0x10; // Length byte 2
		data[3] = (byte) 0xD3; // Structured field id byte 1
		data[4] = (byte) 0xA9; // Structured field id byte 2
		data[5] = (byte) 0x9B; // Structured field id byte 3
		data[6] = 0x00; // Flags
		data[7] = 0x00; // Reserved
		data[8] = 0x00; // Reserved

		for (int i = 0; i < _nameBytes.length; i++) {

			data[9 + i] = _nameBytes[i];

		}

		return data;

	}

	/**
	 * A control sequence is a sequence of bytes that specifies a control
	 * function. A control sequence consists of a control sequence introducer 
	 * and zero or more parameters. The control sequence can extend multiple
	 * presentation text data objects, but must eventually be terminated. 
	 * This method terminates the control sequence.
	 */
	public void endControlSequence() {

		if (currentPresentationTextData == null) {
			startPresentationTextData();
		}

		try {

			currentPresentationTextData.endControlSequence();

		} catch (MaximumSizeExceededException msee) {

			endPresentationTextData();
			endControlSequence();

		}

	}

}